# Specification of the Web API for the Video-AG Website (v0.66).

## Introduction

In this document we use a lot of URLs. Parameters which need to be replaced are put in `{}`. Anything inside and the
braces need to be replaced with the actual argument. These are called _URL Parameters_.

The Base-URL is `/api/{version}`, where `{version}` must be a version id from below. All URLs in this document are
relative to this Base-URL (Unless explicitly
stated otherwise).

The versions are:

| Version | id | Status         |
|---------|----|----------------|
| 0       | v0 | In Development |

HTTP Bodies, if present, are always JSON objects unless explicitly stated otherwise. We use tables to represent the json
object structure:

| Field     | Type    | Notes                    |
|-----------|---------|--------------------------|
| id        | int     |                          |
| something | ?string | We don't know what it is |

Example json object:
<pre>
{
    "id": 42,
    "something": "asdfghjkl"
}
</pre>

The _Field_ is the name of the JSON Key. _Type_ is the type of data. The type may have some additional indicators:

* A question mark indicates that the value is optional.
* An exclamation mark indicates that the value may only be present if the user is a moderator.
* A plus indicates that the field is [directly modifiable](#directly-modifiable-fields)

Some routes, marked as such, require a CSRF token. This token is set as a cookie `csrf_token` by
`POST /authentication/fsmpi` or `POST /authentication/status` and must be included as the `X-Csrf-Token` HTTP Header in
such requests.

On success, status code `200 OK` is always returned unless otherwise noted. Response Body is always present. For
status codes `4xx` and `5xx` the response contains an [api_error](#api_error) (May not be present if the error
happened in a proxy, etc.).

The changelog can be found [at the end](#changelog).

### Site status

The current site status can be retrieved with `GET /status`. It can be in three states

* `available`: The site is running and all routes can be used
* `readonly`: The site cannot update any data. You can only call routes for which is specified:
  * 'This may be used while the site is [read-only](#site-status).'
* `unavailable`: The site is unavailable and the only route you can call is `GET /status`


### Stored Objects

The API is based on a concept of stored objects. In their serialized json form these look exactly the same as non-stored 
objects. However, stored objects have an id which is contained in their serialized form. With this ID the configuration
of an object can be fetched and modified. The configuration of an object lists the object's fields along with their types
and the current values. This allows the configuration to dynamically add or remove fields without the fields being hardcoded.
Some of these fields correspond directly to fields in the object's serialized form. Such fields are marked with a `+` in
this documentation and can also be modified directly. Note that some restrictions from the 
[field description](#field_description) may still apply, and that the configuration should be queried before editing.
Some fields of the serialized json are **not** included in the object's configuration and vice-versa.

Currently, the following objects have a configuration:
* announcement
* chapter
* course
* featured
* lecture
* media_process_template
* publish_medium
* source_medium


TODO parents
TODO deleted
TODO variants

The types of these fields can be any of the [standard types](#Types), any or an array of such values. 
Right now the following types may appear as the type of a field:
* boolean
* course
* datetime
* enum
* int
* lecture
* media_process
* string
* user[]
* view_permissions


For stored objects, a configuration's field does not return/expect a json object but the id of the object.

## Routes

#### `POST /authentication/password`

This may be used while the site is [read-only](#site-status).

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>lecture_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>password</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td>username</td>
        <td>string</td>
        <td></td>
    </tr>
</tbody>
</table>


If the specified lecture does not exist or does not have password authentication, an [api error](#api_error) is returned.

###### Response:


`200 OK` if the username and password are correct for the specified lecture. Otherwise, an [api error](#api_error) (Usually `authentication_failed`).

The username and password are saved, so that different lectures with the same login data are authenticated as well.



---
#### `POST /authentication/fsmpi`

This may be used while the site is [read-only](#site-status).

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>lecture_id</td>
        <td>?int</td>
        <td>Id of lecture for which to get authentication status</td>
    </tr>
    <tr>
        <td>password</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td>username</td>
        <td>string</td>
        <td></td>
    </tr>
</tbody>
</table>


If the specified lecture does not exist, an [api error](#api_error) is returned.

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>is_lecture_authenticated</td>
        <td>?boolean</td>
        <td>Only present if lecture_id was present in request</td>
    </tr>
    <tr>
        <td>user</td>
        <td>!user</td>
        <td>Only present if the user has moderator access</td>
    </tr>
</tbody>
</table>


`200 OK` if login was successful. Otherwise, an [api error](#api_error) (Usually `authentication_failed`).


---
#### `POST /authentication/start_oauth`

This may be used while the site is [read-only](#site-status).

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>type</td>
        <td>string</td>
        <td>Possible values: <code>rwth</code>, <code>moodle</code></td>
    </tr>
</tbody>
</table>


###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>verification_url</td>
        <td>string</td>
        <td>URL at which user must authenticate themselves</td>
    </tr>
</tbody>
</table>



`202 ACCEPTED` with body as described below.

After the user has authenticated themselves, the authentication needs to be finished by requesting the
status: `/authentication/status`.



---
#### `POST /authentication/status`

This may be used while the site is [read-only](#site-status).

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>lecture_id</td>
        <td>?int</td>
        <td>Id of lecture for which to get authentication status</td>
    </tr>
</tbody>
</table>


If the specified lecture does not exist, an [api error](#api_error) is returned.

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>authenticated_methods</td>
        <td>string[]</td>
        <td>Possible values: <code>fsmpi</code>, <code>rwth</code>, <code>moodle</code>, <code>password</code></td>
    </tr>
    <tr>
        <td>in_progress_authentication</td>
        <td>?string</td>
        <td>Possible values: <code>rwth</code>, <code>moodle</code></td>
    </tr>
    <tr>
        <td>is_lecture_authenticated</td>
        <td>?boolean</td>
        <td>Only present if lecture_id was present in request</td>
    </tr>
    <tr>
        <td>user</td>
        <td>!user</td>
        <td></td>
    </tr>
</tbody>
</table>



Also finishes the last authentication started with `/authentication/start`. If that authentication is not yet complete,
`in_progress_authentication` will be set (Such an authentication cannot 'fail', only be successful or timeout).



---
#### `POST /authentication/moderator_logout`

This may be used while the site is [read-only](#site-status).

###### Response:


Only logs moderators out.

`200 OK` if now logged out



---
#### `GET /courses`

This may be used while the site is [read-only](#site-status).

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>courses</td>
        <td>course{}</td>
        <td>Key is course id. The courses **do not** include their lectures</td>
    </tr>
</tbody>
</table>



---
#### `GET /course/{course_id}`
`GET /course/{course_handle}`

This may be used while the site is [read-only](#site-status).

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>course_handle</td>
        <td>string</td>
        <td>Unique handle</td>
    </tr>
    <tr>
        <td>course_id</td>
        <td>int</td>
        <td>Unique id</td>
    </tr>
    <tr>
        <td>include_lectures</td>
        <td>?boolean</td>
        <td>If <code>true</code>, <code>lectures</code> will be present. Default is <code>false</code></td>
    </tr>
</tbody>
</table>


###### Response:

Returns course


---
#### `GET /lecture/{lecture_id}`

This may be used while the site is [read-only](#site-status).

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>course</td>
        <td>course</td>
        <td>The course **does not** include all its lectures</td>
    </tr>
    <tr>
        <td>lecture</td>
        <td>lecture</td>
        <td>Includes chapters and media</td>
    </tr>
</tbody>
</table>



---
#### `PUT /lecture/{lecture_id}/chapter_suggestion`

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>lecture_id</td>
        <td>int</td>
        <td></td>
    </tr>
</tbody>
</table>


<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>name</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td>start_time</td>
        <td>int</td>
        <td>In seconds</td>
    </tr>
</tbody>
</table>


###### Response:

`201 CREATED` if suggestion was saved or an [api_error](#api_error).


---
#### `GET /search`

This may be used while the site is [read-only](#site-status).

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>q</td>
        <td>string</td>
        <td>The term to search for. May not be empty (only whitespace) or longer than 1024 characters or more than 20 words</td>
    </tr>
</tbody>
</table>


###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>course_context</td>
        <td>course{}</td>
        <td>Key is id of course. This contains all courses which are referenced in _courses_ or _lectures_. They **do not** include their lectures.</td>
    </tr>
    <tr>
        <td>courses</td>
        <td>int[]</td>
        <td>Contains the ids of found courses</td>
    </tr>
    <tr>
        <td>lectures</td>
        <td>lecture[]</td>
        <td>Includes chapters and media</td>
    </tr>
</tbody>
</table>



---
#### `PUT /feedback/new`

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>email</td>
        <td>?string</td>
        <td>May not be longer than 128 characters. Must match <code>\S+@\S+\.\S+</code> (More sophisticated checks are allowed, though)</td>
    </tr>
    <tr>
        <td>text</td>
        <td>string</td>
        <td>May not be longer than 16384 characters</td>
    </tr>
</tbody>
</table>


###### Response:

`201 CREATED` if the feedback was saved. Otherwise, an [api error](#api_error).


---
#### `GET /jobs`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>entries_per_page</td>
        <td>?int</td>
        <td>Must be between 10 and 100</td>
    </tr>
    <tr>
        <td>page</td>
        <td>?int</td>
        <td>Zero-indexed. Must not be negative. If this is bigger than the page count, an empty page is returned</td>
    </tr>
    <tr>
        <td>status</td>
        <td>?string</td>
        <td>Only return jobs with this status. Must be one of ['ready', 'spawning', 'running', 'finished', 'finished_and_processed', 'failed', 'failed_and_processed', 'canceled']</td>
    </tr>
    <tr>
        <td>type</td>
        <td>?string</td>
        <td>Only return jobs of this type. May be any string</td>
    </tr>
</tbody>
</table>


###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>page</td>
        <td>job[]</td>
        <td></td>
    </tr>
    <tr>
        <td>page_count</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>user_context</td>
        <td>user{}</td>
        <td>Key is user id. Contains all the users referenced in the jobs</td>
    </tr>
</tbody>
</table>



---
#### `GET /job/{job_id}`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>job_context</td>
        <td>job[]</td>
        <td>Key is job id</td>
    </tr>
    <tr>
        <td>job_id</td>
        <td>int</td>
        <td></td>
    </tr>
</tbody>
</table>



---
#### `GET /source_media`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>entries_per_page</td>
        <td>?int</td>
        <td>Must be between 10 and 100</td>
    </tr>
    <tr>
        <td>include_sorted</td>
        <td>boolean</td>
        <td>Default is false</td>
    </tr>
    <tr>
        <td>page</td>
        <td>?int</td>
        <td>Zero-indexed. Must not be negative. If this is bigger than the page count, an empty page is returned</td>
    </tr>
</tbody>
</table>


###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>course_context</td>
        <td>course{}</td>
        <td>Key is course id</td>
    </tr>
    <tr>
        <td>lecture_context</td>
        <td>lecture{}</td>
        <td>Key is lecture id</td>
    </tr>
    <tr>
        <td>page</td>
        <td>source_medium[]</td>
        <td></td>
    </tr>
    <tr>
        <td>page_count</td>
        <td>int</td>
        <td></td>
    </tr>
</tbody>
</table>



---
#### `POST /run_source_file_sorter`

This may only be used by a moderator and requires the CSRF Token.

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>scheduled</td>
        <td>boolean</td>
        <td>If false, another sorter is probably still running</td>
    </tr>
</tbody>
</table>



---
#### `GET /media_process_templates`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>templates</td>
        <td>media_process_template[]</td>
        <td></td>
    </tr>
</tbody>
</table>



---
#### `GET /lecture/{lecture_id}/media_process_overview`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>publish_media</td>
        <td>publish_medium{}</td>
        <td>Key is id. Note that these contain their full target media as an object and not just the id (So there is some duplicate data)</td>
    </tr>
    <tr>
        <td>source_media</td>
        <td>source_medium{}</td>
        <td>Key is id</td>
    </tr>
    <tr>
        <td>status</td>
        <td>string</td>
        <td>String with some information about the process (e.g. error messages)</td>
    </tr>
    <tr>
        <td>target_media</td>
        <td>target_medium{}</td>
        <td>Key is id. Target media are serialized in extended mode</td>
    </tr>
</tbody>
</table>



---
#### `POST /lecture/{lecture_id}/run_media_process_scheduler`

This may only be used by a moderator and requires the CSRF Token.

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>scheduled</td>
        <td>boolean</td>
        <td>If false, another scheduler is probably still running</td>
    </tr>
</tbody>
</table>



---
#### `GET /status`

This may be used while the site is [read-only or disabled](#site-status).

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>announcements</td>
        <td>announcement[]</td>
        <td></td>
    </tr>
    <tr>
        <td>git_commit_hash</td>
        <td>?string</td>
        <td></td>
    </tr>
    <tr>
        <td>is_debug</td>
        <td>boolean</td>
        <td></td>
    </tr>
    <tr>
        <td>server_name</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td>status</td>
        <td>string</td>
        <td>possible values: <code>available</code>, <code>readonly</code>, <code>unavailable</code></td>
    </tr>
</tbody>
</table>



---
#### `GET /is_running`

This may be used while the site is [read-only or disabled](#site-status).

###### Response:


Returns `200 OK` if the site is ready to answer requests, otherwise `503 SERVICE UNAVAILABLE`. The status may still be
`unavailable` if this return `200 OK`!



---
#### `GET /homepage`

This may be used while the site is [read-only](#site-status).

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>course_context</td>
        <td>course{}</td>
        <td>Key is course id. Contains all courses referenced in the lectures</td>
    </tr>
    <tr>
        <td>featured</td>
        <td>featured[]</td>
        <td></td>
    </tr>
    <tr>
        <td>latest_lectures</td>
        <td>lecture[]</td>
        <td>Latest is first</td>
    </tr>
    <tr>
        <td>upcoming_lectures</td>
        <td>lecture[]</td>
        <td>Soonest is first</td>
    </tr>
</tbody>
</table>



---
#### `GET /object_management/{object_type}/{object_id}/configuration`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>object_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>object_type</td>
        <td>string</td>
        <td>Name of object (as used in this document) which has a [configuration](#Stored-Objects)</td>
    </tr>
</tbody>
</table>


###### Response:

Returns object_configuration


---
#### `PATCH /object_management/{object_type}/{object_id}/configuration`

This may only be used by a moderator and requires the CSRF Token.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>object_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>object_type</td>
        <td>string</td>
        <td>Name of object (as used in this document) which has a [configuration](#Stored-Objects)</td>
    </tr>
</tbody>
</table>


<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>expected_current_values</td>
        <td>{}</td>
        <td>Same format as <code>updates</code>. An error is thrown if the current value is not the value specified here for some field. Values are optional</td>
    </tr>
    <tr>
        <td>updates</td>
        <td>{}</td>
        <td>Key is <code>id_string</code> of field (Provided by the configuration or of a [directly modifiable field](#Stored-Objects)). Value has type of the field</td>
    </tr>
</tbody>
</table>


###### Response:

Updates the specified fields or returns an [api error](#api_error) if the update failed.


---
#### `GET /object_management/{object_type}/new/configuration`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>object_type</td>
        <td>string</td>
        <td>Name of object (as used in this document) which has a [configuration](#Stored-Objects)</td>
    </tr>
</tbody>
</table>


###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>fields</td>
        <td>field_description[]</td>
        <td></td>
    </tr>
    <tr>
        <td>variant_fields</td>
        <td>{}</td>
        <td>If present, a variant for the object must be chosen. Possible variants are keys of map. Values are arrays as <code>fields</code>. Additional fields for that variant are specified in the array</td>
    </tr>
</tbody>
</table>



Returns the fields which may need to be configured to create a new object. For the creation all returned fields **may**
be specified. All fields which do not have a default value **must** be specified.



---
#### `PUT /object_management/{object_type}/new`

This may only be used by a moderator and requires the CSRF Token.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>object_type</td>
        <td>string</td>
        <td>Name of object (as used in this document) which has a [configuration](#Stored-Objects)</td>
    </tr>
</tbody>
</table>


<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>parent_id</td>
        <td>?int</td>
        <td>Id of the parent object. Must be present if, and only if, <code>parent_type</code> is present</td>
    </tr>
    <tr>
        <td>parent_type</td>
        <td>?string</td>
        <td>Type of parent object. Must be present if object can have parent. See [configuration](#Stored-Objects)</td>
    </tr>
    <tr>
        <td>values</td>
        <td>{}</td>
        <td>Key is for field (Provided by the creation configuration; for the specified object type if one is set). Value has type of the field</td>
    </tr>
    <tr>
        <td>variant</td>
        <td>?string</td>
        <td>Must be present if <code>GET /{object_type}/new/configuration</code> included <code>variant_fields</code>. Value must be a key of that map</td>
    </tr>
</tbody>
</table>



For some objects it is guaranteed that they have no variant and that the following fields are sufficient to create
these objects (No other creation fields apart from these, have no default value):

For `lecture` (With types as specified in the `lecture` object):
* `title`
* `location`
* `time`
* `duration`

For `chapter` (With types as specified in the `chapter` object):
* `start_time`
* `name`
* `visible`


###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>id</td>
        <td>int</td>
        <td>Id of the newly created object</td>
    </tr>
</tbody>
</table>


`201 CREATED` if the object was created successfully, otherwise an [api error](#api_error).


---
#### `POST /object_management/modify_many`

This may only be used by a moderator and requires the CSRF Token.

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>objects</td>
        <td>object_update[]</td>
        <td>An object may not occur multiple times. Maximum is 32 objects</td>
    </tr>
</tbody>
</table>


object_update:
<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>expected_current_values</td>
        <td>{}</td>
        <td>Same format as <code>expected_current_values</code> in <code>PATCH /object_management/{object_type}/{object_id}/configuration</code></td>
    </tr>
    <tr>
        <td>id</td>
        <td>int</td>
        <td>Id of the object to update</td>
    </tr>
    <tr>
        <td>type</td>
        <td>string</td>
        <td>Name of object (as used in this document) which has a [configuration](#Stored-Objects)</td>
    </tr>
    <tr>
        <td>updates</td>
        <td>{}</td>
        <td>Same format as <code>updates</code> in <code>PATCH /object_management/{object_type}/{object_id}/configuration</code></td>
    </tr>
</tbody>
</table>


###### Response:

Atomically updates the specified objects or returns an [api error](#api_error) if the update failed.


---
#### `DELETE /object_management/{object_type}/{object_id}`

This may only be used by a moderator and requires the CSRF Token.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>object_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>object_type</td>
        <td>string</td>
        <td>Name of object (as used in this document) which [may be deleted](#Stored-Objects)</td>
    </tr>
</tbody>
</table>


###### Response:

Deletes specified object or returns an [api error](#api_error) if the deletion failed (Does not exist, already deleted).


---
#### `POST /object_management/{object_type}/{object_id}/resurrect`

This may only be used by a moderator and requires the CSRF Token.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>object_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>object_type</td>
        <td>string</td>
        <td>Name of object (as used in this document) which [may be deleted](#Stored-Objects)</td>
    </tr>
</tbody>
</table>


###### Response:

Restores the specified object, which must have been deleted previously, or returns an [api error](#api_error) if it could not be restored (Does not exist, not deleted).


---
#### `GET /object_management/changelog`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>entries_per_page</td>
        <td>?int</td>
        <td>Must be between 10 and 500</td>
    </tr>
    <tr>
        <td>field_id</td>
        <td>?string</td>
        <td>Only return entries for fields with this id. Accepts any string (for legacy objects)</td>
    </tr>
    <tr>
        <td>object_id</td>
        <td>?int</td>
        <td>Only return entries for objects with this id</td>
    </tr>
    <tr>
        <td>object_type</td>
        <td>?string</td>
        <td>Only return entries for objects of this type. Accepts any string (for legacy objects)</td>
    </tr>
    <tr>
        <td>page</td>
        <td>?int</td>
        <td>Zero-indexed. Must not be negative. If this is bigger than the page count, an empty page is returned</td>
    </tr>
    <tr>
        <td>user_id</td>
        <td>?int</td>
        <td>Only return entries for this user</td>
    </tr>
</tbody>
</table>


###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>page</td>
        <td>changelog_entry[]</td>
        <td></td>
    </tr>
    <tr>
        <td>page_count</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>user_context</td>
        <td>user{}</td>
        <td>Key is user id</td>
    </tr>
</tbody>
</table>



---
#### `GET /users`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

###### Response:

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>users</td>
        <td>user[]</td>
        <td></td>
    </tr>
</tbody>
</table>


Returns a list of all users (moderators)


---
#### `GET /user/me/settings`

This may be used while the site is [read-only](#site-status).

This may only be used by a moderator.

###### Response:

Returns settings


---
#### `PATCH /user/me/settings`

This may only be used by a moderator and requires the CSRF Token.

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>updates</td>
        <td>{}</td>
        <td>Key is id of setting_entry from <code>GET /user/me/settings</code> to change (All types except <code>subsection</code>). Value is new value of corresponding type</td>
    </tr>
</tbody>
</table>


###### Response:

Updates the specified settings or returns an [api_error](#api_error) if the update failed.


---
#### `GET /lecture_import/rwth_online/{rwth_course_id}`

This may only be used by a moderator and requires the CSRF Token.

<table>
<thead>
    <th>URL Parameter</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>rwth_course_id</td>
        <td>int</td>
        <td>ID of the course to fetch from RWTHonline</td>
    </tr>
</tbody>
</table>


###### Response:


Queries all events for the specified course from RWTHonline in ICAL format.

Content Type: `text/calendar`



---


## Resources

### Types

| Type                     | Notes                                                                                                                                                                                                 |
|--------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| int                      |                                                                                                                                                                                                       |
| boolean                  |                                                                                                                                                                                                       |
| string                   |                                                                                                                                                                                                       |
| datetime                 | A string which must be in format `YYYY-MM-ddTHH:mm:ss` (Year, Month, day, Hour, minute, second) (Note that this is a local date time according to [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601)) |

Additionally, the following objects may appear as the type of some field:

* `view_permissions`

### Objects

#### announcement

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>expiration_time</td>
        <td>!datetime</td>
        <td></td>
    </tr>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>page_visibility</td>
        <td>enum</td>
        <td></td>
    </tr>
    <tr>
        <td>publish_time</td>
        <td>!datetime</td>
        <td></td>
    </tr>
    <tr>
        <td>text</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>type</td>
        <td>enum</td>
        <td></td>
    </tr>
    <tr>
        <td>visible</td>
        <td>+!boolean</td>
        <td></td>
    </tr>
</tbody>
</table>

#### changelog_entry

<table>
<thead>
    <th>Variant</th>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td rowspan="6">any</td>
        <td>change_time</td>
        <td>datetime</td>
        <td></td>
    </tr>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>modifying_user_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>object_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>object_type</td>
        <td>string</td>
        <td>Note that this object type may not exist anymore</td>
    </tr>
    <tr>
        <td>type</td>
        <td>enum</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="3"><code>creation</code></td>
        <td>parent_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>parent_type</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td>variant</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="1"><code>deletion_change</code></td>
        <td>is_now_deleted</td>
        <td>boolean</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="3"><code>modification</code></td>
        <td>field_id</td>
        <td>string</td>
        <td>Note that this field may not exist anymore</td>
    </tr>
    <tr>
        <td>new_value</td>
        <td>json</td>
        <td>Note that this may contain arbitrary json due to legacy values</td>
    </tr>
    <tr>
        <td>old_value</td>
        <td>?json</td>
        <td>Note that this may contain arbitrary json due to legacy values. Not present if this was at creation</td>
    </tr>
    <tr>
        <td rowspan="1"><code>unknown</code></td>
        <td>data</td>
        <td>json</td>
        <td></td>
    </tr>
</tbody>
</table>

#### chapter

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>name</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>start_time</td>
        <td>+int</td>
        <td>In seconds, >= 0</td>
    </tr>
    <tr>
        <td>visible</td>
        <td>+!boolean</td>
        <td></td>
    </tr>
</tbody>
</table>

#### course

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>allow_embed</td>
        <td>boolean</td>
        <td></td>
    </tr>
    <tr>
        <td>authentication_information</td>
        <td>string</td>
        <td>May contain additional information on how the user can authenticate themselves</td>
    </tr>
    <tr>
        <td>default_authentication_methods</td>
        <td>string[]</td>
        <td>Default authentication methods for lectures in this course. Some Lectures might have different methods (A lecture object always contains its authentication methods). See lecture for possible values</td>
    </tr>
    <tr>
        <td>description</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>full_name</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>handle</td>
        <td>string</td>
        <td>Unique</td>
    </tr>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>internal_comment</td>
        <td>+!string</td>
        <td></td>
    </tr>
    <tr>
        <td>lectures</td>
        <td>?lecture[]</td>
        <td>Only present if requested. All (public) lectures. Sorted by their time ascending. Lectures will include chapters and publish media</td>
    </tr>
    <tr>
        <td>listed</td>
        <td>!boolean</td>
        <td>If false, course can only be visited with its url (Unless user is moderator)</td>
    </tr>
    <tr>
        <td>organizer</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>semester</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>short_name</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>show_chapters_on_course</td>
        <td>boolean</td>
        <td>If true, chapters should be shown on the course page</td>
    </tr>
    <tr>
        <td>topic</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>visible</td>
        <td>+!boolean</td>
        <td></td>
    </tr>
</tbody>
</table>

#### featured

<table>
<thead>
    <th>Variant</th>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td rowspan="6">any</td>
        <td>display_priority</td>
        <td>+!int</td>
        <td>Smallest value is at top. If the order is the same, the lower id is higher</td>
    </tr>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>text</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>title</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>type</td>
        <td>enum</td>
        <td>Specifies the variant</td>
    </tr>
    <tr>
        <td>visible</td>
        <td>+!boolean</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="1"><code>course</code></td>
        <td>course_id</td>
        <td>int</td>
        <td>May be null (deleted/not visible). Does not include lectures</td>
    </tr>
    <tr>
        <td rowspan="1"><code>image</code></td>
        <td>image_url</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="1"><code>lecture</code></td>
        <td>lecture</td>
        <td>lecture</td>
        <td>May be null (deleted/not visible). Does not include chapters and media_sources</td>
    </tr>
    <tr><td><code>plain</code></td></tr>
</tbody>
</table>

#### field_description

<table>
<thead>
    <th>Variant</th>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td rowspan="3">any</td>
        <td>default_value</td>
        <td>?according to <code>type</code></td>
        <td>If present, this is a default value which may be used for this field (Field may be omitted at creation then)</td>
    </tr>
    <tr>
        <td>id</td>
        <td>string</td>
        <td>Lang keys <code>object.{id}.name</code> and <code>object.{id}.tooltip</code> (Tooltip may not be present)</td>
    </tr>
    <tr>
        <td>type</td>
        <td>string</td>
        <td>See [Stored Objects](#Stored-Objects)</td>
    </tr>
    <tr>
        <td rowspan="1">any stored object (not arrays)</td>
        <td>may_be_null</td>
        <td>boolean</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="1"><code>datetime</code></td>
        <td>may_be_null</td>
        <td>boolean</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="1"><code>enum</code></td>
        <td>enums</td>
        <td>string[]</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="2"><code>int</code></td>
        <td>max_value</td>
        <td>?int</td>
        <td>inclusive, >= <code>min_value</code></td>
    </tr>
    <tr>
        <td>min_value</td>
        <td>?int</td>
        <td>inclusive</td>
    </tr>
    <tr>
        <td rowspan="3"><code>string</code></td>
        <td>max_length</td>
        <td>?int</td>
        <td>inclusive, >= <code>min_length</code></td>
    </tr>
    <tr>
        <td>min_length</td>
        <td>?int</td>
        <td>inclusive, >= 0</td>
    </tr>
    <tr>
        <td>pattern</td>
        <td>?string</td>
        <td>A regex pattern which the string needs to match</td>
    </tr>
</tbody>
</table>

#### field_value

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>description</td>
        <td>field_description</td>
        <td></td>
    </tr>
    <tr>
        <td>value</td>
        <td>According to <code>description</code>.<code>type</code></td>
        <td></td>
    </tr>
</tbody>
</table>

#### job

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>cause_job_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>cause_user_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>creation_time</td>
        <td>datetime</td>
        <td></td>
    </tr>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>input_data</td>
        <td>json</td>
        <td></td>
    </tr>
    <tr>
        <td>on_end_event_data</td>
        <td>json</td>
        <td></td>
    </tr>
    <tr>
        <td>on_end_event_type</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td>output_data</td>
        <td>json</td>
        <td></td>
    </tr>
    <tr>
        <td>run_end_time</td>
        <td>datetime</td>
        <td></td>
    </tr>
    <tr>
        <td>run_start_time</td>
        <td>datetime</td>
        <td></td>
    </tr>
    <tr>
        <td>status</td>
        <td>enum</td>
        <td></td>
    </tr>
    <tr>
        <td>type</td>
        <td>string</td>
        <td></td>
    </tr>
</tbody>
</table>

#### lecture

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>authentication_methods</td>
        <td>string[]</td>
        <td>possible values for string: <code>public</code> (Everyone has access, array will contain no other values), <code>rwth</code>, <code>moodle</code>, <code>fsmpi</code>, <code>password</code>; Empty array indicates no access</td>
    </tr>
    <tr>
        <td>chapters</td>
        <td>?chapter[]</td>
        <td>Only present if requested. Sorted according to the start_time (ascending order)</td>
    </tr>
    <tr>
        <td>course_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>description</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>duration</td>
        <td>+int</td>
        <td>In minutes</td>
    </tr>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>internal_comment</td>
        <td>+!string</td>
        <td></td>
    </tr>
    <tr>
        <td>livestream_planned</td>
        <td>boolean</td>
        <td></td>
    </tr>
    <tr>
        <td>location</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>no_recording</td>
        <td>+boolean</td>
        <td></td>
    </tr>
    <tr>
        <td>publish_media</td>
        <td>?publish_medium[]</td>
        <td>Only present if requested. All (public) media. May be empty</td>
    </tr>
    <tr>
        <td>publish_time</td>
        <td>!datetime</td>
        <td></td>
    </tr>
    <tr>
        <td>speaker</td>
        <td>+string</td>
        <td>Name of the speaker</td>
    </tr>
    <tr>
        <td>time</td>
        <td>+datetime</td>
        <td></td>
    </tr>
    <tr>
        <td>title</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>visible</td>
        <td>+!boolean</td>
        <td></td>
    </tr>
</tbody>
</table>

#### media_process

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>producers</td>
        <td>media_process_target_producer[]</td>
        <td></td>
    </tr>
    <tr>
        <td>publish_target_ids</td>
        <td>string[]</td>
        <td>Values must be ids of targets produced the the producers. These media are published</td>
    </tr>
    <tr>
        <td>publish_wait_for_full_process</td>
        <td>boolean</td>
        <td>If true, the media are published once the full process has run through. Otherwise a medium is published as soon as it is finished</td>
    </tr>
</tbody>
</table>

#### media_process_target_producer

<table>
<thead>
    <th>Variant</th>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td rowspan="1">any</td>
        <td>type</td>
        <td>string</td>
        <td>See variant of this object for possible values</td>
    </tr>
    <tr>
        <td rowspan="1"><code>NOCHECK: </code>downscale_video<code></code></td>
        <td>source_id</td>
        <td>string</td>
        <td>Id of the input target</td>
    </tr>
    <tr>
        <td rowspan="1"><code>NOCHECK: </code>source_file<code>, </code>sample_thumbnail<code>, </code>downscale_video<code></code></td>
        <td>output_id</td>
        <td>string</td>
        <td>Id of the target to output</td>
    </tr>
    <tr>
        <td rowspan="1"><code>downscale_video</code></td>
        <td>target_vertical_resolution</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="1"><code>sample_thumbnail</code></td>
        <td>timestamp_percent</td>
        <td>int</td>
        <td>Between 0 and 100 (both inclusive). Specifies at which time the thumbnail is sampled</td>
    </tr>
    <tr>
        <td rowspan="1"><code>source_file</code></td>
        <td>tag</td>
        <td>string</td>
        <td>Tags are used to match uploaded files to targets. Tags must match [a-zA-Z0-9_]*. Every file has exactly one tag and this target takes the file with the tag specified here. No tag on a file is equivalent to an empty string tag</td>
    </tr>
</tbody>
</table>

#### media_process_template

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>name</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>process</td>
        <td>+media_process</td>
        <td></td>
    </tr>
</tbody>
</table>

#### object_configuration

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>fields</td>
        <td>int</td>
        <td></td>
    </tr>
</tbody>
</table>

#### publish_medium

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>target_medium</td>
        <td>target_medium</td>
        <td></td>
    </tr>
    <tr>
        <td>title</td>
        <td>+string</td>
        <td></td>
    </tr>
    <tr>
        <td>visible</td>
        <td>+!boolean</td>
        <td></td>
    </tr>
</tbody>
</table>

#### settings

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>sections</td>
        <td>settings_section[]</td>
        <td></td>
    </tr>
</tbody>
</table>

#### settings_entry

<table>
<thead>
    <th>Variant</th>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td rowspan="2">any</td>
        <td>id</td>
        <td>string</td>
        <td>Lang Key: <code>setting.entry.{id}.name</code> and <code>setting.entry.{id}.description</code> (description may not exist)</td>
    </tr>
    <tr>
        <td>type</td>
        <td>string</td>
        <td>Type of this entry. For possible values see this table</td>
    </tr>
    <tr>
        <td rowspan="1"><code>boolean</code></td>
        <td>value</td>
        <td>boolean</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="1"><code>subsection</code></td>
        <td>entries</td>
        <td>settings_entry[]</td>
        <td>Contain no subsections</td>
    </tr>
</tbody>
</table>

#### settings_section

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>entries</td>
        <td>settings_entry[]</td>
        <td></td>
    </tr>
    <tr>
        <td>name</td>
        <td>string</td>
        <td>Lang Key: <code>setting.section.{name}</code></td>
    </tr>
</tbody>
</table>

#### source_medium

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>file_metadata</td>
        <td>json</td>
        <td>Only calculated once this is sorted. There are no guarantees regarding the content</td>
    </tr>
    <tr>
        <td>file_modification_time</td>
        <td>datetime</td>
        <td></td>
    </tr>
    <tr>
        <td>file_path</td>
        <td>string</td>
        <td>Path is relative to website data directory. Does NOT start with a /</td>
    </tr>
    <tr>
        <td>file_size</td>
        <td>int</td>
        <td>Only calculated once this is sorted</td>
    </tr>
    <tr>
        <td>force_immediate_sort</td>
        <td>+boolean</td>
        <td>Set this to true to sort the file immediately (when sorter is running) ignoring a recent modification, file format or any previous error</td>
    </tr>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>lecture_id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>sha256</td>
        <td>string</td>
        <td>Only calculated once this is sorted</td>
    </tr>
    <tr>
        <td>sorter_error_message</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td>status</td>
        <td>enum</td>
        <td></td>
    </tr>
    <tr>
        <td>tag</td>
        <td>string</td>
        <td></td>
    </tr>
    <tr>
        <td>update_time</td>
        <td>datetime</td>
        <td></td>
    </tr>
</tbody>
</table>

#### target_medium

<table>
<thead>
    <th>Variant</th>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td rowspan="8">any</td>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>input_data_sha256</td>
        <td>?string</td>
        <td>Only included in extended mode (explicitly stated)</td>
    </tr>
    <tr>
        <td>is_produced</td>
        <td>?boolean</td>
        <td>Only included in extended mode (explicitly stated)</td>
    </tr>
    <tr>
        <td>process_sha256</td>
        <td>?string</td>
        <td>Only included in extended mode (explicitly stated)</td>
    </tr>
    <tr>
        <td>process_target_id</td>
        <td>?string</td>
        <td>Only included in extended mode (explicitly stated)</td>
    </tr>
    <tr>
        <td>producer_job</td>
        <td>?job</td>
        <td>Only included in extended mode (explicitly stated)</td>
    </tr>
    <tr>
        <td>to_be_replaced</td>
        <td>?boolean</td>
        <td>Only included in extended mode (explicitly stated)</td>
    </tr>
    <tr>
        <td>type</td>
        <td>enum</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="3"><code>image</code></td>
        <td>horizontal_resolution</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>url</td>
        <td>string</td>
        <td>URL to the mediums's file (Maybe with a redirect)</td>
    </tr>
    <tr>
        <td>vertical_resolution</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="4"><code>plain_audio</code></td>
        <td>audio_channel_count</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>audio_sample_rate</td>
        <td>int</td>
        <td>In Hz</td>
    </tr>
    <tr>
        <td>duration_sec</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>url</td>
        <td>string</td>
        <td>URL to the mediums's file (Maybe with a redirect)</td>
    </tr>
    <tr>
        <td rowspan="8"><code>plain_video</code></td>
        <td>audio_channel_count</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>audio_sample_rate</td>
        <td>int</td>
        <td>In Hz</td>
    </tr>
    <tr>
        <td>duration_sec</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>horizontal_resolution</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>url</td>
        <td>string</td>
        <td>URL to the mediums's file (Maybe with a redirect)</td>
    </tr>
    <tr>
        <td>vertical_resolution</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>video_frame_rate_denominator</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>video_frame_rate_numerator</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td rowspan="3"><code>thumbnail</code></td>
        <td>horizontal_resolution</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>url</td>
        <td>string</td>
        <td>URL to the mediums's file (Maybe with a redirect)</td>
    </tr>
    <tr>
        <td>vertical_resolution</td>
        <td>int</td>
        <td></td>
    </tr>
</tbody>
</table>

#### user

<table>
<thead>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td>handle</td>
        <td>!string</td>
        <td></td>
    </tr>
    <tr>
        <td>id</td>
        <td>int</td>
        <td></td>
    </tr>
    <tr>
        <td>name</td>
        <td>!string</td>
        <td></td>
    </tr>
</tbody>
</table>

#### view_permissions

<table>
<thead>
    <th>Variant</th>
    <th>Field</th>
    <th>Type</th>
    <th>Notes</th>
</thead>
<tbody>
    <tr>
        <td rowspan="1">any</td>
        <td>type</td>
        <td>string</td>
        <td>Possible values: <code>inherit</code>, <code>public</code>, <code>private</code>, <code>authentication</code>. For type <code>authentication</code> at least one authentication method must be allowed/specified</td>
    </tr>
    <tr>
        <td rowspan="4"><code>authentication</code></td>
        <td>fsmpi_authentication</td>
        <td>boolean</td>
        <td></td>
    </tr>
    <tr>
        <td>moodle_course_ids</td>
        <td>int[]</td>
        <td></td>
    </tr>
    <tr>
        <td>passwords</td>
        <td>string{}</td>
        <td>Key is username. Value is password. Maximum string length: 1024</td>
    </tr>
    <tr>
        <td>rwth_authentication</td>
        <td>boolean</td>
        <td></td>
    </tr>
</tbody>
</table>



#### api_error

| Field      | Type      | Notes                                       |
|------------|-----------|---------------------------------------------|
| error_code | id_string | Possible values are listed below            |
| message    | string    | Human readable message                      |
| server     | string    | Name of the server which served the request |

Possible `error_code`:

| Code                                  | HTTP Code | Notes                                                |
|---------------------------------------|-----------|------------------------------------------------------|
| bad_request                           | 400       |                                                      |
| method_not_allowed                    | 405       |                                                      |
| malformed_json                        | 400       | Content-Type needs to be set to `application/json`   |
| request_missing_parameter             | 400       |                                                      |
| request_invalid_parameter             | 400       |                                                      |
| unknown_request_path                  | 400       |                                                      |
| internal_server_error                 | 500       |                                                      |
| lecture_has_no_password               | 400       |                                                      |
| authentication_failed                 | 403       |                                                      |
| authentication_not_available          | 503       |                                                      |
| unauthorized                          | 401       |                                                      |
| csrf_token_invalid                    | 403       | CSRF token must be included as `X-Csrf-Token` Header |
| access_forbidden                      | 403       |                                                      |
| rate_limited                          | 429       |                                                      |
| unknown_object                        | 404       |                                                      |
| object_error                          | 400       |                                                      |
| modification_unexpected_current_value | 409       |                                                      |
| too_many_suggestions                  | 403       |                                                      |
| site_is_readonly                      | 503       |                                                      |
| site_is_disabled                      | 503       |                                                      |
| site_is_overloaded                    | 503       |                                                      |

## Changelog

### v0.66

* `GET /source_media` is only for mods
* `GET /lecture/{lecture_id}/media_process_overview` is only for mods
* Added `POST /run_source_file_sorter`
* Added `POST /lecture/{lecture_id}/run_media_process_scheduler`
* Added `GET /media_process_templates`
* `old_value` of `changelog_entry` is no longer present if the entry was at creation
* `publish_time` of `lecture` is now only returned for mods
* Added `name` and `process` to `media_process_template`

### v0.65

* Changes in `target_medium`
  * Renamed `video_horizontal_resolution` to `horizontal_resolution`
  * Renamed `video_vertical_resolution` to `vertical_resolution`
  * Added `horizontal_resolution` and `vertical_resolution` to `thumbnail`
  * Added `image` variant

### v0.64

* Added `GET /source_media`
* Fixed `source_id` name of `media_process_target_producer` (Was documented as `input_id`)
* Added `sample_thumbnail` variant to `media_process_target_producer`
* Changes in `source_medium`
  * Added `force_immediate_sort`
  * Added `lecture_id`
  * Added `status`
  * Added `tag`
* Added `GET /lecture/{lecture_id}/media_process_overview`
* Changes in `target_medium`
  * Added `input_data_sha256`
  * Added `is_produced`
  * Added `process_sha256`
  * Added `process_target_id`
  * Added `producer_job`
  * Added `to_be_replaced`

### v0.63

* Added `file_metadata`, `file_modification_time`, `file_size` to `source_medium`
* Added and renamed variables in `target_medium`

### v0.62

* Removed `thumbnail_url` from `lecture`
* Added new variant `thumbnail` to `target_medium`
* Removed `visible` from `source_medium` (Only accessible by mods anyway)
* Added missing variant `plain` to `featured` and `plain_audio` to `target_medium` which were mentioned nowhere before

### v0.61

* Renamed `path` to `file_path` in `source_medium`
* Added `update_time` to `source_medium`

### v0.60

* In `job`
  * Renamed `run_finish_time` to `run_end_time` (to avoid confusing with `finished` status. This time is also set if failed)
  * Added `on_end_event_data` and `on_end_event_type`
* In `source_medium`
  * Added `path`, `sha256`, `sorter_error_message`, `sorting_paused_due_to_recent_modification`
* Completely redid `media_process`
* Renamed `media_process_target` to `media_process_target_producer`
  * Removed `id`
  * Added `output_id`
  * Changed `tags` list to a single `tag`

### v0.59

* Changed `entries_per_page` maximum in `GET /jobs` to 100
* Changed response of `GET /jobs`
* Removed `GET /events`
* Removed `event`
* Removed `cause_event_id` from `job`
* Added `cause_user_id` to `job`
* Added `GET /job/{job_id}`

### v0.58

* Renamed `frame_rate_nominator` to `frame_rate_numerator` in `target_medium`

### v0.57

* Removed `GET /diagnostics`
* Added `thumbnail_url` to `lecture`
* Added `url` to `target_medium`

### v0.56

* Added stored objects `media_process_template`, `target_medium` and `source_medium`
* Added non-stored object `media_process`, `media_process_target`
* Added `GET /jobs`
* Added `GET /events`
* Added `user_context` to `GET /object_management/changelog`
* Added `target_medium` to `publish_medium`

### v0.55

* Removed all 'types' except `int`, `boolean`, `string`, `datetime`
  * Most other types were just some of the standard data types with some additional conditions/notes and only used once. 
    These notes are now just wherever the type is used
* Changed allowed fields for configuration objects
* Changes in `GET /search`
  * Renamed `courses_context` to `course_context`
  * The search term may not have more than 20 words
* Removed `GET /feedback?entries_per_page={entries_per_page}&page={page}`, `feedback_entry`
* Changes in `GET /object_management/changelog`
  * Renamed URL parameter `field` to `field_id`
  * URL parameter `field_id` may now be set even if the `object_type` is missing
  * Added URL parameter `user_id`
* Removed `GET /sorter/log`, `DELETE /sorter/error_log/{log_id}`, `sorter_log_entry`, `sorter_error_log_entry`
* Removed `GET /jobs`, `DELETE /job/{id}`, `POST /job/{id}/action`, `job`, `worker`
* Removed `announcement_configuration`, `featured_configuration`, `course_configuration`, `lecture_configuration`
  * Added `object_configuration` instead (Format is exactly the same)
* Removed `homepage` object. This is now an anonymous response object of `GET /homepage`
  * Renamed `courses_context` to `course_context`
  * Renamed `new_lectures` to `latest_lectures`
* Changes in `announcement`
  * Removed `is_currently_visible`, `has_expired`
  * Added `publish_time`, `expiration_time`, `visible`
  * Renamed `visibility` to `page_visibility`
* Changes in `changelog_entry`
  * `object_type` and `object_id` are now for all variants
  * `object_type` may be for type which no longer exists (due to legacy values)
  * `new_value` and `old_value` no longer have types guaranteed to match the field's type (due to legacy values)
  * All fields for type `unknown` have been removed and a new json field `data` has been added
* Changes in `chapter`
  * Renamed `is_visible` to `visible`
* Changes in `course`
  * Renamed `is_visible` to `visible`
  * Renamed `is_listed` to `listed`
  * Renamed `id_string` to `handle`
  * `topic` is now directly modifiable
  * `visible` is now directly modifiable
  * Added `allow_embed`
  * Added `internal_comment`
* Changes in `featured`
  * `id` is now also returned if non-mod
  * Removed guarantees for `display_priority`. These might be added back later
  * Removed `course_list` type
  * Added `course` type
  * Renamed `is_visible` to `visible`
* Changes in `field_description`
  * Removed fields for `long`, `url`, `datetime`, `lecture_id`, `user_id`
  * Added `enums`
* Changes in `lecture`
  * Removed `thumbnail_url` (Will be added back later)
  * Removed `livestream_url`
  * Removed `allow_embed` (The course now has this field)
  * Renamed `media_sources` to `publish_media`
  * Renamed `is_visible` to `visible`
* Changes in `media_source` which is now named `publish_medium`
  * `id` is now also returned for non-mods
  * Removed `quality`, `size`, `player_url`, `download_url`. These will be added back later
  * Removed `internal_comment`
  * Renamed `is_visible` to `visible`
  * Renamed `comment` to `title`
* Removed `media_quality`
* Removed `feedback_entry`
* Changes in `user_info` which is now named `user`
  * Renamed `id_string` to `handle`

### v0.54

* Added `PUT /feedback/new`, `GET /feedback?entries_per_page={entries_per_page}&page={page}`
* Added `feedback_entry` object

### v0.53

* Changed type of `course_id_string` to regex. Added forbidden values.

### v0.52

* Added error code `site_is_overloaded`

### v0.51

* Added `server_name` to `GET /status`

### v0.50

* Removed `GET /announcements`
* Updated `GET /status`. Among other things, this now returns the announcements
* Changed name of error type `authentication_servers_not_available` to `authentication_not_available`
* Added error types `site_is_readonly` and `site_is_disabled`
* Added `GET /is_running`

### v0.49

* Added `GET /status`

### v0.48

* Removed `PUT /job/new/{type}`
* Removed `POST /sorter/run`

### v0.47

* Removed `GET /object_management/{object_type}/{object_id}/{field_id}/changelog`
* Removed `GET /object_management/{object_type}/{object_id}/changelog`
* Added query parameters to `GET /object_management/changelog` instead
* Update type `unknown` of `changelog_entry`:
  * `unknown_type` -> `unknown_object_type` 
  * `old_value` -> `param1` (May be null)
  * `new_value` -> `param2` 

### v0.46

* Type `url` may be type of field.

### v0.45

* Removed `permission` object
* Removed `permissions` from the configurations of courses and lectures
* Added `view_permissions` object which may appear as the type of some field
* `may_be_null` may now only appear in `field_description` of type `datetime`, `lecture_id` or `user_id`
* `duration` and `timestamp` may no longer appear as the type of some field

### v0.44

* Added `expected_current_values` to `PATCH /object_management/{object_type}/{object_id}/configuration` (and accordingly to `POST /object_management/modify_many`)
* Added `modification_unexpected_current_value` error code

### v0.43

* Added `parent_type` and `parent_id` to `changelog_entry` of type `creation`

### v0.42

* Added `worker` object. Added `workers` field to `GET /jobs`
* Removed `POST /job/{id}/restart`, `POST /job/{id}/cancel` and `POST /job/{id}/copy`
* Added `POST /job/{id}/action` instead

### v0.41

* Added `GET /jobs`, `PUT /job/new/{type}`, `DELETE /job/{id}`, `POST /job/{id}/restart`, `POST /job/{id}/cancel` and `POST /job/{id}/copy`
* Added `job` object
* Added `job_state` and `job_type` types

### v0.40

* Added `GET /user/me/settings` and `PATCH /user/me/settings`
* Added `settings`, `settings_section` and `settings_entry` object

### v0.39

* Added `sorter_log_entry` and `sorter_error_log_entry` objects
* Added `GET /sorter/log`, `DELETE /sorter/error_log/{log_id}` and `POST /sorter/run`

### v0.38

* Removed `is_modifiable` of `field_description` (All fields can be modified)
* Fields with default values may be omitted at creation now
* Added guarantees for the creation of some objects

### v0.37

* CSRF token is now returned in a cookie instead of the response body 

### v0.36

* Added CSRF tokens. The token is returned by `POST /authentication/status`.
* CSRF token is required by:
  - `GET /lecture_import/rwth_online/{rwth_course_id}`
  - `PATCH /object_management/{object_type}/{object_id}/configuration`
  - `POST /object_management/modify_many`
  - `PUT /object_management/{object_type}/new`
  - `DELETE /object_management/{object_type}/{object_id}`
  - `POST /object_management/{object_type}/{object_id}/resurrect`
* Added error code `csrf_token_invalid`

### v0.35

* Added `POST /object_management/{object_type}/{object_id}/resurrect`

### v0.34

* Added error code `rate_limited`

### v0.33

* Removed `field_entry`
* Added `field_description` and `field_value` instead. The format of the configuration endpoints and `changelog_entry` has changed accordingly. 

### v0.32

* Changes to [changelog_entry](#changelog_entry) format
* Added `GET /object_management/{object_type}/{object_id}/changelog?entries_per_page={entries_per_page}&page={page}`

### v0.31

* Changed `GET /{object_type}/{object_id}/configuration` to `GET /object_management/{object_type}/{object_id}/configuration`
* Changed `PATCH /{object_type}/{object_id}/configuration` to `PATCH /object_management/{object_type}/{object_id}/configuration`
* Changed `POST /objects/modify_many` to `POST /object_management/modify_many`
* Changed `GET /{object_type}/new/configuration` to `GET /object_management/{object_type}/new/configuration`
* Changed `PUT /{object_type}/new/create` to `PUT /object_management/{object_type}/new`
* Changed `DELETE /{object_type}/{object_id}` to `DELETE /object_management/{object_type}/{object_id}`
* Changed `GET /{object_type}/{object_id}/{field_id}/changelog` to `GET /object_management/{object_type}/{object_id}/{field_id}/changelog`
* Changed `GET /objects/changelog` to `GET /object_management/changelog`

### v0.30

* Updated `GET /{object_type}/{object_id}/{field_id}/changelog`
* Added `GET /search?q={search_term}`

### v0.29

* Added `GET /{object_type}/{object_id}/{field_id}/changelog`
* Changed `/lecture_import/rwthonline/{rwth_course_id}` to ``/lecture_import/rwth_online/{rwth_course_id}``

### v0.28

* `/lecture_import/list_document/{course_id}` => `/lecture_import/rwthonline/{rwth_course_id}`
* Removed `POST /lecture_import/calendar_export`

### v0.27

* Added `GET /lecture_import/list_document/{course_id}`
* Added `POST /lecture_import/calendar_export`

### v0.26

* Added `GET /objects/changelog?entries_per_page={entries_per_page}&page={page}`

### v0.25

* Allowed creation of `chapter` objects
* Added `PUT /lecture/{lecture_id}/chapter_suggestion`
* Added `too_many_suggestions` error type

### v0.24

* Added `allow_embed` to `lecture`

### v0.23

* Updated directly modifiable Fields. They are now indicated by a `+`
* Added `legacy_announcement_type`

### v0.22

* Changed featured type `text` to `plain` (Because all types contain text)

### v0.21

* Updated `display_priority` of `featured` to be unique and no longer nullable

### v0.20

* Added `DELETE /{object_type}/{object_id}`

### v0.19

* Changed type of `courses` in `featured` to array of int. Added courses to the `courses_context` of the homepage instead.

### v0.18

* Renamed Object Types to Object Variants as this was confusing. Renamed `type_fields` to `variant_fields` and `type` to `variant` for this

### v0.17

* Added `POST /objects/modify_many`

### v0.16

* Added `display_priority` to `featured` (Also to [Directly Modifiable Fields](#Directly-Modifiable-Fields))

### v0.15

* Added `course_filter_criteria` type
* `id_string` may no longer be type of field entry
* Added `course_id_string` type

### v0.14

* Added `GET /users`

### v0.13

* Reverted `GET /announcement` back to `GET /announcements`
* Reverted `GET /course` back to `GET /courses`

### v0.12

* Added `POST /authentication/moderator_logout`

### v0.11

* Added Object Modification API
* Added `announcement_configuration`, `featured_configuration`, `course_configuration`, `lecture_configuration`, `permission` and `field_entry` objects
* Added `internal_comment` to `media_source`
* Added `unknown_object` and `object_error` error types. Removed `unknown_course` and `unknown_lecture` error types.
* Changed `GET /announcements` to `GET /announcement` (for consistency)
* Changed `GET /courses` to `GET /course` (for consistency)
* Changed `GET /courses/{course_id}` to `GET /course/{course_id}` (for consistency)
* Changed `GET /lectures/{lecture_id}` to `GET /lecture/{lecture_id}` (for consistency)

### v0.10

* Added `id`, `is_currently_visible` and `has_expired` to `announcement` object
* Added `id` and `is_visible` to `featured` object
* Added `is_listed` and `is_visible` to `course` object
* Added `is_visible` and `internal_comment` to `lecture` object
* Added `id` and `is_visible` to `chapter` object
* Added `id` and `is_visible` to `media_source` object
* Changed pattern for `id_string` from `[a-z0-9_-]{1,100}` to `(?=.*[a-z_-])[a-z0-9_-]{1,100}`
* Changed route `GET /courses/{course_id}/lectures/{lecture_id}` to `GET /lectures/{lecture_id}`
* Made `lecture` of `featured` nullable 

### v0.9

* Added `user_info` to response of `POST /authentication/status`

### v0.8

* Added `authentication_information` to `course` object

### v0.7

* Added `bad_request` and `method_not_allowed` error codes

### v0.6

* Changed `/authentication/start` to `/authentication/start_oauth`
* Removed `/authentication/login`
* Added `/authentication/password` and `/authentication/fsmpi` instead

### v0.5

* Changed `moderator` authentication type to `fsmpi` (To be consistent with `authentication_methods`).

### v0.3

* `GET /courses/{course_id}/lectures/{lecture_id}` now accepts course id_string and includes the course

### v0.2

* `GET /courses` returns map instead of array

### v0.1

* Initial version
