import secrets
import re
import string
from ipaddress import ip_address, ip_network
import requests
from requests.exceptions import JSONDecodeError
from functools import wraps
from flask import request

from videoag_common.objects import *

from api.session import api_session
from api.miscellaneous import *
from api.database import *
import api

_BASE_URL = api.config["API_BASE_URL"]


def api_moderator_route(require_csrf_token: bool = False):
    def decorator(func):
        if hasattr(func, "api_function"):
            raise Exception("@api_moderator_route() seems to be applied after @api_route(). @api_moderator_route() "
                            "should be the first (lowest) decorator!")
        
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not is_moderator():
                raise ApiClientException(ERROR_UNAUTHORIZED)
            if require_csrf_token:
                _check_csrf_token()
            return func(*args, **kwargs)
        
        setattr(wrapper, "api_function_only_mod", True)
        setattr(wrapper, "api_function_requires_csrf_token", require_csrf_token)
        return wrapper
    
    return decorator


def is_moderator():
    return api_session.get_user_id() is not None


def _check_csrf_token():
    if "X-Csrf-Token" not in request.headers \
            or not is_valid_csrf_token(request.headers["X-Csrf-Token"]):
        raise ApiClientException(ERROR_INVALID_CSRF_TOKEN)


def is_valid_csrf_token(csrf_token: str) -> bool:
    return api_session.csrf_token is not None and api_session.csrf_token == csrf_token


def get_csrf_token():
    if api_session.csrf_token is None:
        raise RuntimeError("No csrf token in session. Probably not a moderator")
    return api_session.csrf_token


def get_user_id() -> int or None:
    return api_session.get_user_id()


def get_user_info() -> dict or None:
    return api_session.get_user_info()


def is_rwth_ip(ip_string: str) -> bool:
    ip = ip_address(ip_string)
    for net in api.config['RWTH_IP_RANGES']:
        if ip in ip_network(net):
            return True
    return False


def is_lecture_authenticated(lecture_id: int) -> bool:
    if is_moderator():
        return True
    lecture = database.query_one_or_none_and_expunge(
        Lecture.select(is_mod=False, to_load=[LOAD_LECTURE_COURSE], visibility_check=True)
               .where(Lecture.id == lecture_id)
    )
    if lecture is None:
        raise ApiClientException(ERROR_OBJECT_ERROR(f"Unknown lecture with id '{lecture_id}'"))
    return is_authenticated(lecture.effective_view_permissions)


def is_authenticated(perm: EffectiveViewPermissions):
    if is_moderator():
        return True
    match perm.type:
        case ViewPermissionsType.PUBLIC:
            return True
        case ViewPermissionsType.PRIVATE:
            return False
        case ViewPermissionsType.AUTHENTICATION:
            if perm.rwth_authentication and (
                    api_session.is_rwth_authenticated
                    or ("X-Real-IP" in request.headers and is_rwth_ip(request.headers["X-Real-IP"]))
            ):
                return True
            
            if perm.fsmpi_authentication and is_moderator():
                return True
            
            for moodle_id in perm.moodle_course_ids:
                if api_session.is_in_moodle_course(moodle_id):
                    return True
            
            for username, password in perm.passwords.items():
                if api_session.is_password_authenticated(username, password):
                    return True
        
        case ViewPermissionsType.INHERIT:
            raise ValueError("Got INHERIT permission from EffectiveViewPermissions")
        case _:
            raise ValueError(f"Unknown type {perm.type}")


def authenticate_password(lecture_id: int, username: str, password: str):
    """
    May throw APIClientException.
    
    Only returns if authentication was successful.
    """
    
    lecture = database.query_one_or_none_and_expunge(
        Lecture.select(is_mod=False, to_load=[LOAD_LECTURE_COURSE], visibility_check=True)
        .where(Lecture.id == lecture_id)
    )
    if lecture is None:
        raise ApiClientException(ERROR_OBJECT_ERROR(f"Unknown lecture with id '{lecture_id}'"))
    perm = lecture.effective_view_permissions
    
    if perm.type != ViewPermissionsType.AUTHENTICATION or len(perm.passwords) == 0:
        raise ApiClientException(ERROR_LECTURE_HAS_NO_PASSWORD)
    
    if username not in perm.passwords or password != perm.passwords[username]:
        raise ApiClientException(ERROR_AUTHENTICATION_FAILED)
    
    api_session.add_authenticated_password(username, password)


if DEBUG_ENABLED:
    def __ldap_authenticate(user: str, password: str) -> tuple[str, str, str, list[str]] or None:
        if user == "videoag" and password == "videoag":
            return "videoag", "Video", "Video", ["fachschaft", "videoag"]
        return None
elif "LDAP_HOST" in api.config:
    import ldap3
    from ldap3.core.exceptions import LDAPBindError, LDAPPasswordIsMandatoryError
    
    __LDAP_USER_REGEX = re.compile("[a-z0-9]+")
    __LDAP_MEMBER_OF_REGEX = re.compile("CN=([a-zA-Z0-9_-]+),CN=Users,DC=fsmpi,DC=rwth-aachen,DC=de")
    __LDAP_HOST = api.config["LDAP_HOST"]
    __LDAP_PORT = api.config["LDAP_PORT"]
    
    
    def __ldap_authenticate(user: str, password: str) -> tuple[str, str, str, list[str]] or None:
        """
        Returns user id, given name, surname, list of groups
        """
        if __LDAP_USER_REGEX.fullmatch(user) is None:
            return None
        try:
            server = ldap3.Server(host=__LDAP_HOST, port=__LDAP_PORT, use_ssl=True)
            conn = ldap3.Connection(server, f"fsmpi\\{user}", password, auto_bind=True, check_names=False)
            conn.search("cn=users,dc=fsmpi,dc=rwth-aachen,dc=de",
                        f"(cn={user})",
                        attributes=["memberOf", "givenName", "sn"])
            attributes = conn.response[0]["attributes"]
            
            # TODO There also exists a display name
            given_name = attributes["givenName"][0]  # Attribute should have exactly one value
            surname = attributes["sn"][0]  # Attribute should have exactly one value
            groups = []
            for val in attributes["memberOf"]:
                match = __LDAP_MEMBER_OF_REGEX.fullmatch(val)
                if not match:
                    continue
                groups.append(match.group(1))
            
            return user, given_name, surname, groups
        except (LDAPBindError, LDAPPasswordIsMandatoryError):
            return None
else:
    def __ldap_authenticate(user: str, password: str) -> tuple[str, str, str, list[str]] or None:
        raise ApiClientException(ERROR_AUTHENTICATION_NOT_AVAILABLE())


def authenticate_fsmpi(username: str, password: str) -> {}:
    """
    May throw APIClientException.
    
    Only returns if authentication was successful.
    """
    user_data = __ldap_authenticate(username, password)
    if user_data is None:
        raise ApiClientException(ERROR_AUTHENTICATION_FAILED)
    user_handle, given_name, surname, groups = user_data
    if not __ldap_is_moderator(groups):
        raise ApiClientException(ERROR_AUTHENTICATION_FAILED)
    
    if api.live_config.is_readonly():
        user_db = database.query_one_or_none_and_expunge(User.select(is_mod=True, to_load=[]).where(User.handle == user_handle))
        if user_db is None:
            raise ApiClientException(ERROR_AUTHENTICATION_NOT_AVAILABLE(
                "Site is read-only and we can not create a new account for you in the database"))
    else:
        def _trans(session: SessionDb):
            user_db = session.scalar(User.select(is_mod=True, to_load=[]).where(User.handle == user_handle))
            if user_db is None:
                # TODO test new
                user_db = User(handle=user_handle, name=given_name)
                session.add(user_db)
                session.flush() # Get autoincrement id

            user_db.last_login = datetime.now()
            session.expunge_all()
            session.commit()
            return user_db
        user_db = database.execute_write_transaction(_trans)
    
    api_session.set_user(user_db)
    api_session.csrf_token = "".join(
        secrets.choice(string.ascii_letters + string.digits) for _ in range(64))
    api_session.permanent = True
    
    return api_session.get_user_info()


def logout_moderator():
    api_session.set_user(None)
    api_session.csrf_token = None
    api_session.permanent = False


def __ldap_is_moderator(user_groups: []) -> bool:
    for group in api.config["LDAP_GROUPS"]:
        if group in user_groups:
            return True
    return False


def start_rwth_authentication() -> str:
    """
    :return: Verification url
    """
    return __start_oauth("rwth")


def is_rwth_authentication_running() -> bool:
    return api_session.running_oauth_scope == "rwth"


def is_moodle_authentication_running() -> bool:
    return api_session.running_oauth_scope == "moodle"


def start_moodle_authentication() -> str:
    """
    :return: Verification url
    """
    return __start_oauth("moodle")


if DEBUG_ENABLED:
    
    debug_running_oauth = {}
    
    def __start_oauth(scope: str) -> str:
        if scope not in ["rwth", "moodle"]:
            raise ValueError("Expected scope to be rwth or moodle")
        oauth_code = "".join(
            secrets.choice(string.ascii_letters + string.digits) for _ in range(64))
        
        debug_running_oauth[oauth_code] = {
            "scope": scope
        }
        api_session.running_oauth_scope = scope
        api_session.running_oauth_code = oauth_code
        return _BASE_URL + "/debug/dummy_oauth/auth/" + oauth_code
    
    
    def try_finish_running_authentication():
        if api_session.running_oauth_scope is None or api_session.running_oauth_code is None:
            return
        oauth_scope = api_session.running_oauth_scope
        oauth_code = api_session.running_oauth_code
        
        running_oauth = debug_running_oauth.get(oauth_code)
        if running_oauth is None or not running_oauth.get("finished", False):
            return
        
        api_session.running_oauth_scope = None
        api_session.running_oauth_code = None
        
        api_session.is_rwth_authenticated = True
        if oauth_scope == "moodle":
            if "moodle_courses" in running_oauth:
                api_session.set_moodle_course_ids(running_oauth["moodle_courses"])
            else:
                api_session.set_moodle_course_ids([])
        
        return
    
    from api.version import API_LATEST_VERSION, get_api_path
    
    @api.app.route(get_api_path(API_LATEST_VERSION, "/debug/dummy_oauth/auth/<code>"), methods=["GET"])
    def debug_authentication_dummy_oauth_auth(code):
        if code not in debug_running_oauth:
            return "Unknown code", HTTP_404_NOT_FOUND
        if debug_running_oauth[code].get("scope") == "moodle":
            return f"""
                <html>
                    <h1>Dummy Moodle OAuth</h1>
                    <form action="{get_api_path(API_LATEST_VERSION, '/debug/dummy_oauth/submit/' + code)}" method="post">
                        <label for="moodle_courses">Moodle courses (ids comma separated):</label>
                        <input type="text" name="moodle_courses" id="moodle_courses"/>
                        <input type="submit"/>
                    </form>
                </html>
                """, HTTP_200_OK
        else:
            return f"""
                <html>
                    <h1>Dummy RWTH OAuth</h1>
                    <form action="{get_api_path(API_LATEST_VERSION, '/debug/dummy_oauth/submit/' + code)}" method="post">
                        <input type="submit"/>
                    </form>
                </html>
                """, HTTP_200_OK
    
    
    @api.app.route(get_api_path(API_LATEST_VERSION, "/debug/dummy_oauth/submit/<code>"), methods=["POST"])
    def debug_authentication_dummy_oauth_submit(code):
        if code not in debug_running_oauth:
            return "Unknown code", HTTP_404_NOT_FOUND
        
        running_oauth = debug_running_oauth[code]
        running_oauth["finished"] = True
        
        if running_oauth.get("scope") == "moodle" and "moodle_courses" in request.form:
            courses_str: str = request.form["moodle_courses"]
            if len(courses_str.strip()) > 0:
                running_oauth["moodle_courses"] = [int(s) for s in courses_str
                                                   .replace("\r", "")
                                                   .replace("\n", "")
                                                   .replace(" ", "")
                                                   .replace("\t", "")
                                                   .split(",")]
        return "Successfully Authenticated", HTTP_200_OK

else:
    def __start_oauth(scope: str) -> str:
        if scope not in ["rwth", "moodle"]:
            raise ValueError("Expected scope to be rwth or moodle")
        response = __make_oauth_request(
            "code",
            scope="moodle.rwth" if scope == "moodle" else "userinfo.rwth"
        )
        api_session.running_oauth_code = response["device_code"]
        api_session.running_oauth_scope = scope
        return response["verification_url"] + "?q=verify&d=" + response["user_code"]
    
    
    def try_finish_running_authentication():
        if api_session.running_oauth_code is None or api_session.running_oauth_scope is None:
            return
        oauth_scope = api_session.running_oauth_scope
        if oauth_scope not in ["rwth", "moodle"]:
            raise ValueError("Session auth scope set to something different than rwth or moodle")
        try:
            token_response = __make_oauth_request("token", code=api_session.running_oauth_code, grant_type="device")
            if token_response.get("status") != "ok":
                return
            api_session.running_oauth_code = None
            api_session.running_oauth_scope = None
            
            api_session.is_rwth_authenticated = True
            if oauth_scope == "moodle":
                moodle_response = __make_moodle_request("getmyenrolledcourses", token_response["access_token"])
                if moodle_response and moodle_response.get("Data"):
                    api_session.set_moodle_course_ids(
                        map(lambda course: int(course["id"]), moodle_response["Data"])
                    )
                else:
                    # TODO notify
                    pass
            
            __make_oauth_request("token", refresh_token=token_response["refresh_token"], grant_type="invalidate")
        except ApiClientException:
            pass


def get_currently_authenticated_methods() -> [str]:
    methods = []
    if api_session.is_rwth_authenticated:
        methods.append("rwth")
    if api_session.is_in_any_moodle_course():
        methods.append("moodle")
    if api_session.has_any_password_authenticated():
        methods.append("password")
    if is_moderator():
        methods.append("fsmpi")
    return methods


OAUTH_BASE = "https://oauth.campus.rwth-aachen.de/oauth2waitress/oauth2.svc/"
MOODLE_BASE = "https://moped.ecampus.rwth-aachen.de/proxy/api/v2/eLearning/Moodle/"


def __make_moodle_request(endpoint, token, **args):
    args["token"] = token
    return __handle_request_result(requests.request('GET', MOODLE_BASE + endpoint, params=args, timeout=30))


def __make_oauth_request(endpoint, **args):
    if "RWTH_API_KEY" not in api.config:
        raise ApiClientException(ERROR_AUTHENTICATION_NOT_AVAILABLE())
    
    args["client_id"] = api.config["RWTH_API_KEY"]
    return __handle_request_result(requests.request('POST', OAUTH_BASE + endpoint, data=args, timeout=30))


def __handle_request_result(response: requests.Response):
    if response.ok:
        return response.json()
    
    try:
        response = response.json()
    except JSONDecodeError:
        response = None
        pass
    
    if response.status_code >= 500:
        raise ApiClientException(ERROR_AUTHENTICATION_NOT_AVAILABLE())
    raise Exception("Error during request. Status: " + str(response.status_code) + " Response: " + str(response))
