import os
import traceback
from pathlib import Path

import api
from api.miscellaneous import *

_LIVE_CONFIG_PATH_STRING = os.environ.get("VIDEOAG_API_LIVE_CONFIG", None)
_LIVE_CONFIG_PATH = None if _LIVE_CONFIG_PATH_STRING is None else Path(_LIVE_CONFIG_PATH_STRING).resolve()


class LiveConfig:
    
    def __init__(self):
        super().__init__()
        self._disabled: bool = False
        self._readonly: bool = False
        self._static_announcements: list[dict] = []
    
    def reset(self):
        self._disabled = False
        self._readonly = False
        self._static_announcements = []
    
    def update(self, json_config: CJsonObject):
        """
        Raises an exception if the file has wrong format/values. In this case, nothing is updated
        """
        disable = json_config.get_bool("disabled", False)
        readonly = json_config.get_bool("readonly", False) or disable
        static_announcements = []
        unsafe_announcements = json_config.get_array("static_announcements")
        if unsafe_announcements is not None:
            for unsafe_announcement in unsafe_announcements:
                unsafe_announcement = unsafe_announcement.as_object()
                
                type = unsafe_announcement.get_string("type", 100)
                if type not in ["info", "warning", "important"]:
                    unsafe_announcement.get("type").raise_error("Unknown type")
                visibility = unsafe_announcement.get_string("visibility", 100)
                if visibility not in ["only_main_page", "all_pages", "all_pages_and_embed"]:
                    unsafe_announcement.get("visibility").raise_error("Unknown visibility")
                text = unsafe_announcement.get_string("text", 8192)
                
                if not unsafe_announcement.get_bool("is_enabled", True):
                    continue
                static_announcements.append({
                    "type": type,
                    "visibility": visibility,
                    "text": text
                })
        
        # Assign to global vars only if whole file parsed successfully
        self._readonly = readonly
        self._disabled = disable
        self._static_announcements = static_announcements
    
    def is_disabled(self) -> bool:
        return self._disabled
    
    def is_readonly(self) -> bool:
        return self._readonly
    
    def get_static_announcements(self) -> list[dict]:
        """
        Dicts contain (as in the api specification): type, visibility, text
        """
        return self._static_announcements


# Do not modify directly! Variable is imported by others
config: LiveConfig = LiveConfig()


@scheduled_function(api.config["LIVE_CONFIG_UPDATE_INTERVAL_SECONDS"], api.config["LIVE_CONFIG_UPDATE_INTERVAL_SECONDS"])
def _update_live_config():
    if _LIVE_CONFIG_PATH is None:
        return
    if not _LIVE_CONFIG_PATH.exists():
        config.reset()
        return
    config.update(CJsonObject(json.loads(_LIVE_CONFIG_PATH.read_text(encoding="UTF-8"))))
    # Exception is caught and logged by scheduler


try:
    _update_live_config()
except Exception as e:
    print(f"Exception when loading live config for the first time:")
    traceback.print_exception(e)
