from abc import ABC, ABCMeta, abstractmethod
from threading import Lock

import api
from api.miscellaneous.scheduler import scheduled_function


class DiagnosticsCounter:
    def __init__(self, key: str, tracker: "DiagnosticsTracker"):
        """Do not use this constructor! Use :func:`DiagnosticsTracker.register_counter`"""
        super().__init__()
        self._key = key
        self._tracker = tracker
    
    def trigger(self):
        # noinspection PyProtectedMember
        self._tracker._trigger_counter(self._key)


class DiagnosticsDataProvider(ABC):
    __metaclass__ = ABCMeta
    
    @abstractmethod
    def get_diagnostics_data(self) -> dict[str, int or str]:
        pass  # pragma: no cover


class DiagnosticsTracker:
    
    def __init__(self, max_saved_interval_count: int):
        super().__init__()
        self._max_saved_interval_count = max_saved_interval_count
        
        self._lock = Lock()
        self._history_lock = Lock()
        self._counters_template: dict[str, int] = {}
        self._providers: list[tuple[str, DiagnosticsDataProvider]] = []
        
        self._history: list[dict[str, int]] = []
        
        self._current_counters: dict[str, int] = {}
    
    def register_counter(self, key: str) -> DiagnosticsCounter:
        with self._lock:
            if key in self._counters_template:
                raise ValueError(f"A counter with key {key} already exists")  # pragma: no cover
            self._counters_template[key] = 0
            self._current_counters[key] = 0
            return DiagnosticsCounter(key, self)
    
    def register_provider(self, key_prefix: str, provider: DiagnosticsDataProvider):
        with self._lock:
            self._providers.append((key_prefix, provider))
    
    def _trigger_counter(self, key: str):
        with self._lock:
            if key not in self._current_counters:
                raise ValueError("Unknown counter")  # pragma: no cover
            self._current_counters[key] += 1
    
    def end_interval(self):
        with self._lock:
            data = self._current_counters
            self._current_counters = self._counters_template.copy()
            
            for key_prefix, provider in self._providers:
                provider_data = provider.get_diagnostics_data()
                for sub_key, value in provider_data.items():
                    key = f"{key_prefix}.{sub_key}"
                    if key in data:
                        print(f"Warning: Duplicate key {key} in diagnostics data")  # pragma: no cover
                    data[key] = value
        
        with self._history_lock:
            if len(self._history) >= self._max_saved_interval_count:
                self._history.pop(0)
            self._history.append(data)
    
    def get_history(self) -> list[dict[str, int]]:
        with self._history_lock:
            return self._history.copy()


_history_size_min = api.config.get("API_DIAGNOSTICS_HISTORY_SIZE_MINUTES", 60)
_interval_size_min = api.config.get("API_DIAGNOSTICS_INTERVAL_SIZE_MINUTES", 60)

if _history_size_min % _interval_size_min != 0:  # pragma: no cover
    print("Warning: Interval size for diagnostics does not fit in history size")
    _history_size_min = _interval_size_min * (_history_size_min // _interval_size_min)

DIAGNOSTICS_TRACKER = DiagnosticsTracker(_history_size_min / _interval_size_min)


@scheduled_function(_interval_size_min * 60, _interval_size_min * 60)
def _diagnostics_tracker_end_interval():
    DIAGNOSTICS_TRACKER.end_interval()
