from flask import request

from api.routes import *
from api.authentication import (is_lecture_authenticated, get_currently_authenticated_methods,
                                authenticate_password, authenticate_fsmpi, logout_moderator,
                                start_rwth_authentication, start_moodle_authentication,
                                is_rwth_authentication_running, is_moodle_authentication_running,
                                try_finish_running_authentication,
                                get_user_info, get_csrf_token)
import api

_API_AUTH_RATE_LIMITERS = create_configured_host_rate_limiters("authentication", api.config["API_AUTH_RATE_LIMIT"])
_COOKIES_SAMESITE = api.config.get("COOKIES_SAMESITE", "strict")
_COOKIES_SECURE = api.config.get("COOKIES_SECURE", True)
_COOKIES_DOMAIN = api.config.get("COOKIES_DOMAIN", None)


@api_add_route("/authentication/password", "POST")
@api_function(rate_limiters=_API_AUTH_RATE_LIMITERS, allow_while_readonly=True,
              request_objects=[
                  ("username", "string"),
                  ("password", "string"),
                  ("lecture_id", "int")
              ],
              description="If the specified lecture does not exist or does not have password authentication, an "
                          "[api error](#api_error) is returned.",
              response_description="""
`200 OK` if the username and password are correct for the specified lecture. Otherwise, an [api error](#api_error) (Usually `authentication_failed`).

The username and password are saved, so that different lectures with the same login data are authenticated as well.
""")
def api_route_authentication_password():
    json_request = get_client_json(request)
    lecture_id: int = json_request.get_sint32("lecture_id",)
    username: str = json_request.get_string("username", max_length=256)
    password: str = json_request.get_string("password", max_length=256)
    authenticate_password(lecture_id, username, password)
    return {}


@api_add_route("/authentication/fsmpi", "POST")
@api_function(rate_limiters=_API_AUTH_RATE_LIMITERS, allow_while_readonly=True,
              request_objects=[
                  ("username", "string"),
                  ("password", "string"),
                  ("lecture_id", "?int", "Id of lecture for which to get authentication status")
              ],
              description="If the specified lecture does not exist, an [api error](#api_error) is returned.",
              response_description="`200 OK` if login was successful. Otherwise, an [api error](#api_error) (Usually "
                                   "`authentication_failed`).",
              response_objects=[
                  ("user", "!user", "Only present if the user has moderator access"),
                  ("is_lecture_authenticated", "?boolean", "Only present if lecture_id was present in request")
              ])
def api_route_authentication_fsmpi():
    json_request = get_client_json(request)
    username: str = json_request.get_string("username", max_length=256)
    password: str = json_request.get_string("password", max_length=256)
    user_info = authenticate_fsmpi(username, password)
    
    json_response = {}
    if user_info is not None:
        json_response["user"] = user_info
    if json_request.has("lecture_id"):
        json_response["is_lecture_authenticated"] = is_lecture_authenticated(json_request.get_sint32("lecture_id"))

    response = ApiResponse(json_response)
    _set_moderator_cookies(response)
    return response


@api_add_route("/authentication/start_oauth", "POST")
@api_function(rate_limiters=_API_AUTH_RATE_LIMITERS, allow_while_readonly=True,
              request_objects=[
                  ("type", "string", "Possible values: `rwth`, `moodle`")
              ],
              response_description="""
`202 ACCEPTED` with body as described below.

After the user has authenticated themselves, the authentication needs to be finished by requesting the
status: `/authentication/status`.
""",
              response_objects=[
                  ("verification_url", "string", "URL at which user must authenticate themselves")
              ])
def api_route_authentication_start_oauth():
    json_request = get_client_json(request)
    login_type: str = json_request.get_string("type", max_length=10)
    if login_type == "rwth":
        verification_url = start_rwth_authentication()
    elif login_type == "moodle":
        verification_url = start_moodle_authentication()
    else:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER("type", "Unknown type"))
    return {
        "verification_url": verification_url
    }, HTTP_202_ACCEPTED


@api_route("/authentication/status", "POST", allow_while_readonly=True,
           request_objects=[
               ("lecture_id", "?int", "Id of lecture for which to get authentication status")
           ],
           description="If the specified lecture does not exist, an [api error](#api_error) is returned.",
           response_objects=[
               ("authenticated_methods", "string[]", "Possible values: `fsmpi`, `rwth`, `moodle`, `password`"),
               ("in_progress_authentication", "?string", "Possible values: `rwth`, `moodle`"),
               ("is_lecture_authenticated", "?boolean", "Only present if lecture_id was present in request"),
               ("user", "!user")
           ],
           response_description="""
Also finishes the last authentication started with `/authentication/start`. If that authentication is not yet complete,
`in_progress_authentication` will be set (Such an authentication cannot 'fail', only be successful or timeout).
""")
def api_route_authentication_status():
    json_request = get_client_json(request)
    
    try_finish_running_authentication()
    
    json_response = {
        "authenticated_methods": get_currently_authenticated_methods()
    }
    
    if is_rwth_authentication_running():
        json_response["in_progress_authentication"] = "rwth"
    elif is_moodle_authentication_running():
        json_response["in_progress_authentication"] = "moodle"
    
    if json_request.has("lecture_id"):
        json_response["is_lecture_authenticated"] = is_lecture_authenticated(json_request.get_sint32("lecture_id"))
    
    if is_moderator():
        json_response["user"] = get_user_info()
    
    response = ApiResponse(json_response)
    _set_moderator_cookies(response)
    return response


def _set_moderator_cookies(response: ApiResponse):
    if is_moderator():
        response.response.set_cookie("csrf_token",
                                     get_csrf_token(),
                                     max_age=None,  # Only for session
                                     httponly=False,
                                     samesite=_COOKIES_SAMESITE,
                                     secure=_COOKIES_SECURE,
                                     domain=_COOKIES_DOMAIN)
        # Used for nginx caching
        response.response.set_cookie("moderator",
                                     "#",
                                     max_age=None,
                                     httponly=True,
                                     samesite=_COOKIES_SAMESITE,
                                     secure=_COOKIES_SECURE,
                                     domain=_COOKIES_DOMAIN)


@api_route("/authentication/moderator_logout", "POST", allow_while_readonly=True,
           response_description="""
Only logs moderators out.

`200 OK` if now logged out
""")
def api_route_authentication_moderator_logout():
    logout_moderator()
    response = ApiResponse({})
    response.response.delete_cookie("csrf_token", httponly=False, samesite=_COOKIES_SAMESITE, secure=_COOKIES_SECURE, domain=_COOKIES_DOMAIN)
    response.response.delete_cookie("moderator", httponly=True, samesite=_COOKIES_SAMESITE, secure=_COOKIES_SECURE, domain=_COOKIES_DOMAIN)
    return response
