from typing import Sequence

from flask import request

from videoag_common.objects import *

from api.routes import *


@api_route("/courses", "GET", allow_while_readonly=True,
           response_objects=[
               ("courses", "course{}", "Key is course id. The courses **do not** include their lectures")
           ])
def api_route_courses():
    is_mod: bool = is_moderator()
    
    courses = database.query_all_and_expunge(
        Course.select(is_mod, to_load=[], visibility_check=True)
              .order_by(Course.id.asc())
    )
    return {"courses": {
        str(course.id): course.serialize(
            is_mod=is_mod,
            include_lectures=False
        ) for course in courses
    }}


@api_add_route("/course/<int:course_id>", "GET")
@api_add_route("/course/<string:course_handle>", "GET")
@api_function(allow_while_readonly=True,
              url_parameters=[
                  ("course_id", "int", "Unique id"),
                  ("course_handle", "string", "Unique handle"),
                  ("include_lectures", "?boolean", "If `true`, `lectures` will be present. Default is `false`")
              ],
              response_object_id="course")
def api_route_course(course_id: int = None, course_handle: str = None):
    is_mod: bool = is_moderator()
    include_lectures = "include_lectures" in request.args and request.args["include_lectures"] == "true"
    
    stmt = (
        Course.select(is_mod, [
            LOAD_COURSE_LECTURES if include_lectures else None,
            LOAD_LECTURE_CHAPTERS,
            LOAD_LECTURE_PUBLISH_MEDIA,
            LOAD_PUBLISH_MEDIUM_TARGET_MEDIUM
        ], visibility_check=False, ignore_unlisted=True)
    )
    
    if course_id:
        stmt = stmt.where(Course.id == course_id)
    else:
        stmt = stmt.where(Course.handle == course_handle)
    
    course = database.query_one_or_none_and_expunge(stmt)
    if not course:
        raise ApiClientException(ERROR_UNKNOWN_OBJECT)
    
    assert isinstance(course, Course)
    if not course.has_access(is_mod, visibility_check=True, ignore_unlisted=True):
        raise ApiClientException(ERROR_UNAUTHORIZED)
    
    return course.serialize(
        is_mod=is_mod,
        include_lectures=include_lectures,
        include_media=True,
        include_chapters=True,
        course_context={str(course.id): None}  # Required because the lecture would serialize its course into the context
    )


@api_route("/lecture/<int:lecture_id>", "GET", allow_while_readonly=True,
           response_objects=[
               ("course", "course", "The course **does not** include all its lectures"),
               ("lecture", "lecture", "Includes chapters and media")
           ])
def api_route_lecture(lecture_id: int):
    is_mod: bool = is_moderator()
    
    stmt = Lecture.select(
        is_mod,
        [
            LOAD_LECTURE_COURSE,
            LOAD_LECTURE_CHAPTERS,
            LOAD_LECTURE_PUBLISH_MEDIA,
            LOAD_PUBLISH_MEDIUM_TARGET_MEDIUM
        ],
        visibility_check=False,
    ).where(Lecture.id == lecture_id)
    
    lecture = database.query_one_or_none_and_expunge(stmt)
    if lecture is None:
        raise ApiClientException(ERROR_UNKNOWN_OBJECT)
    
    assert isinstance(lecture, Lecture)
    if not lecture.has_access(is_mod, visibility_check=True, ignore_unlisted=True):
        raise ApiClientException(ERROR_UNAUTHORIZED)
    
    course_context = {}
    lecture_json = lecture.serialize(
        is_mod=is_mod,
        include_media=True,
        include_chapters=True,
        course_context=course_context,
        include_lectures=False,
    )
    return {
        "course": course_context[str(lecture.course_id)],
        "lecture": lecture_json
    }


_CHAPTER_LIMIT_FOR_SUGGESTIONS = api.config["API_CHAPTER_SUGGESTIONS_LIMIT_PER_LECTURE"]
_CHAPTER_SUGGESTIONS_RATE_LIMITER = IntervalRateLimiter(
    api.config["API_CHAPTER_SUGGESTIONS_RATE_LIMIT"]["window_size_seconds"] * pow(10, 9),
    api.config["API_CHAPTER_SUGGESTIONS_RATE_LIMIT"]["max_request_count"],
    api.config["API_CHAPTER_SUGGESTIONS_RATE_LIMIT"]["interval_size_seconds"] * pow(10, 9),
    "chapter_suggestions.rate_limiter"
)


@api_route("/lecture/<int:lecture_id>/chapter_suggestion", "PUT",
           url_parameters=[
               ("lecture_id", "int")
           ],
           request_objects=[
               ("start_time", "int", "In seconds"),
               ("name", "string")
           ],
           response_description="`201 CREATED` if suggestion was saved or an [api_error](#api_error).")
def api_route_lecture_chapter_suggestion(lecture_id: int):
    is_mod = is_moderator()
    json_request = get_client_json(request)
    start_time = json_request.get_int("start_time", 0, MAX_VALUE_SINT32)
    name = json_request.get_string("name", 128)
    
    if not _CHAPTER_SUGGESTIONS_RATE_LIMITER.check_new_request():
        raise ApiClientException(ERROR_TOO_MANY_SUGGESTIONS)
    
    def _trans(session: SessionDb):
        lecture = session.scalar(
            Lecture.select(
                        is_mod,
                        [
                            LOAD_LECTURE_COURSE,
                            LOAD_LECTURE_CHAPTERS,
                        ],
                        visibility_check=False,
                    )
                   .where(Lecture.id == lecture_id)
        )
        if lecture is None:
            raise ApiClientException(ERROR_UNKNOWN_OBJECT)
        
        assert isinstance(lecture, Lecture)
        if not lecture.has_access(is_mod, visibility_check=True, ignore_unlisted=True):
            raise ApiClientException(ERROR_UNAUTHORIZED)
        
        if len(lecture.chapters) >= _CHAPTER_LIMIT_FOR_SUGGESTIONS:
            raise ApiClientException(ERROR_TOO_MANY_SUGGESTIONS)
        
        lecture.chapters.append(Chapter(
            time=start_time,
            text=name,
            created_by=get_user_id() if is_mod else None
        ))
        session.commit()
    
    database.execute_write_transaction(_trans)
    return {}, HTTP_201_CREATED


def _execute_search(session: SessionDb,
                    id_column: orm.properties.MappedColumn[int],
                    search_columns: list[orm.properties.MappedColumn[str]],
                    base_select: sql.Select,
                    extra_order_by: list,
                    terms: list[str]):
    sub_queries: list[sql.Select] = []
    prio = len(terms)
    for term in terms:
        sub_queries.append(
            sql.select(
                id_column.label("_id"),
                sql.cast(prio, sql.types.Integer).label("_priority")
            )
            .where(
                sql.or_(*map(lambda col: col.ilike(term), search_columns))
            )
        )
        prio -= 1
    
    return session.scalars(
        base_select.join(
            sql.select(
                sql.literal_column("_id", sql.types.Integer),
                sql.func.cast(sql.func.sum(sql.literal_column("_priority", sql.types.Integer)), sql.types.Integer)
                .label("_score")
            )
            .select_from(sql.union_all(*sub_queries).subquery())
            .group_by(sql.literal_column("_id", sql.types.Integer))
            .subquery("_data"),
            id_column == sql.literal_column("_data._id", sql.types.Integer)
        )
        .order_by(sql.literal_column("_data._score", sql.types.Integer).desc(), *extra_order_by)
    ).all()


def _execute_lecture_search(session: SessionDb, is_mod: bool, terms: list[str]) -> Sequence[Lecture]:
    return _execute_search(
        session,
        Lecture.id,
        [Lecture.title, Lecture.description, Lecture.speaker],
        Lecture.select(
                    is_mod,
                    [
                        LOAD_LECTURE_COURSE,
                        LOAD_LECTURE_CHAPTERS,
                        LOAD_LECTURE_PUBLISH_MEDIA,
                        LOAD_PUBLISH_MEDIUM_TARGET_MEDIUM
                    ],
                    visibility_check=True,
                )
               .limit(30),
        [Lecture.time.desc()],
        terms
    )


def _execute_course_search(session: SessionDb, is_mod: bool, terms: list[str]) -> Sequence[Course]:
    return _execute_search(
        session,
        Course.id,
        [Course.full_name, Course.short_name, Course.handle, Course.organizer, Course.topic, Course.description],
        Course.select(is_mod, to_load=[], visibility_check=True)
              .limit(20),
        [Course.semester.desc()],
        terms
    )


def _db_execute_search(session: SessionDb, is_mod: bool, terms: list[str]):
    lectures = _execute_lecture_search(session, is_mod, terms)
    courses = _execute_course_search(session, is_mod, terms)
    session.expunge_all()
    session.rollback()
    return lectures, courses


@api_route("/search", "GET", allow_while_readonly=True,
           url_parameters=[
               ("q", "string", "The term to search for. May not be empty (only whitespace) or longer than 1024"
                               " characters or more than 20 words")
           ],
           response_objects=[
               ("courses", "int[]", "Contains the ids of found courses"),
               ("lectures", "lecture[]", "Includes chapters and media"),
               ("course_context", "course{}", "Key is id of course. This contains all courses which are referenced in "
                                              "_courses_ or _lectures_. They **do not** include their lectures.")
           ])
def api_route_search():
    is_mod = is_moderator()
    
    if "q" not in request.args:
        raise ApiClientException(ERROR_REQUEST_MISSING_PARAMETER("URL.q"))
    search_term = request.args["q"]
    if len(search_term) > 1024:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER("URL.q", "Too long"))
    if search_term.isspace():
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER("URL.q", "Only whitespace is not allowed"))
    
    words: list[str] = list(filter(lambda w: not w.isspace(), search_term.split(" ")))
    if len(words) > 20:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER("URL.q", "Too many search terms"))
    
    terms = list(map(
        lambda w: "%"
                  + w.lower()
                     .replace("%", "\\%")
                     .replace("_", "\\_")
                  + "%",
        words
    ))
    
    lectures, courses = database.execute_read_transaction(_db_execute_search, is_mod, terms)
    
    course_ids = []
    lectures_json = []
    course_context = {}
    
    for lecture in lectures:
        lectures_json.append(lecture.serialize(
            is_mod=is_mod,
            include_chapters=True,
            include_media=True,
            course_context=course_context,
            include_lectures=False
        ))
    
    for course in courses:
        course_ids.append(course.serialize_to_context(
            is_mod=is_mod,
            course_context=course_context,
            include_lectures=False)
        )
    
    return {
        "courses": course_ids,
        "lectures": lectures_json,
        "course_context": course_context
    }
