import math
from collections import defaultdict

from videoag_common.objects import *

from ..routes import *


@api_route("/jobs", "GET", allow_while_readonly=True,
           url_parameters=[
               ("entries_per_page", "?int", "Must be between 10 and 100"),
               ("page", "?int", "Zero-indexed. Must not be negative. If this is bigger than the page count, an empty "
                                "page is returned"),
               ("status", "?string", f"Only return jobs with this status. Must be one of "
                                     f"{list(map(lambda m: m.value, JobState.__members__.values()))}"),
               ("type", "?string", f"Only return jobs of this type. May be any string"),
           ],
           response_objects=[
               ("page_count", "int"),
               ("page", "job[]"),
               ("user_context", "user{}", "Key is user id. Contains all the users referenced in the jobs"),
           ])
@api_moderator_route()
def api_route_get_jobs():
    entries_per_page = api_request_get_query_int("entries_per_page", 20, 10, 50)
    page_num = api_request_get_query_int("page", 0)
    
    status: str or None = api_request_get_query_string("status", 100)
    type: str or None = api_request_get_query_string("type", 100)
    
    def _trans(session: SessionDb):
        query = Job.select(is_mod=True, to_load=[])
        
        # TODO check status, type valid
        if status is not None:
            query = query.where(Job.status == status)
        if type is not None:
            query = query.where(Job.type == type)
        
        entries = session.scalars(
            query.offset(page_num * entries_per_page)
            .options(orm.joinedload(Job.cause_user))
            .limit(entries_per_page)
            .order_by(Job.creation_time.desc())
        ).all()
        count = session.scalar(
            query.with_only_columns(sql.func.count()).select_from(Job)
        )
        
        user_context = {}
        
        page = []
        for entry in entries:
            page.append(entry.serialize(
                is_mod=True,
                user_context=user_context,
                job_context=NoSerializationContext()
            ))
        
        session.rollback()
        return {
            "page_count": math.ceil(count / entries_per_page),
            "page": page,
            "user_context": user_context,
        }
    return database.execute_read_transaction(_trans)


@api_route("/job/<int:job_id>", "GET", allow_while_readonly=True,
           response_objects=[
               ("job_id", "int"),
               ("job_context", "job[]", "Key is job id"),
           ])
@api_moderator_route()
def api_route_get_job_graph(job_id: int):
    
    def _trans(session: SessionDb):
        job = session.scalar(
            Job.select(is_mod=True, to_load=[])
            .where(Job.id == job_id)
            .options(orm.joinedload(Job.cause_user))
        )
        if job is None:
            raise ApiClientException(ERROR_UNKNOWN_OBJECT)
        
        job_context = {}
        user_context = {}
        
        job.serialize_to_context(is_mod=True, job_context=job_context, user_context=user_context)
        
        session.rollback()
        return {
            "job_id": job_id,
            "job_context": job_context,
            "user_context": user_context,
        }
    
    return database.execute_read_transaction(_trans)
