import math
from itertools import chain

from videoag_common.media_process.basic_targets import SOURCE_FILE_TAG_PATTERN

from ..routes import *
from videoag_common.objects import *

api_register_non_stored_object("media_process_target_producer", [
    (None, "type", "string", "See variant of this object for possible values"),
    ("NOCHECK: `source_file`, `sample_thumbnail`, `downscale_video`",
     "output_id", "string", "Id of the target to output"),
    ("NOCHECK: `downscale_video`",
     "source_id", "string", "Id of the input target"),
    ("source_file", "tag", "string", f"Tags are used to match uploaded files to targets. Tags must match "
                                     f"{SOURCE_FILE_TAG_PATTERN.pattern}. Every file has exactly one tag and this target "
                                     f"takes the file with the tag specified here. No tag on a file is equivalent to an "
                                     f"empty string tag"),
    ("downscale_video", "target_vertical_resolution", "int", ""),
    ("sample_thumbnail", "timestamp_percent", "int", "Between 0 and 100 (both inclusive). Specifies at which time the "
                                                     "thumbnail is sampled"),
])

api_register_non_stored_object("media_process", [
    ("producers", "media_process_target_producer[]"),
    ("publish_target_ids", "string[]", "Values must be ids of targets produced the the producers. These media are published"),
    ("publish_wait_for_full_process", "boolean", "If true, the media are published once the full process has run through. "
                                                 "Otherwise a medium is published as soon as it is finished"),
])


@api_route("/source_media", "GET", allow_while_readonly=True,
           url_parameters=[
               ("entries_per_page", "?int", "Must be between 10 and 100"),
               ("page", "?int", "Zero-indexed. Must not be negative. If this is bigger than the page count, an empty "
                                "page is returned"),
               ("include_sorted", "boolean", f"Default is false"),
           ],
           response_objects=[
               ("page_count", "int"),
               ("page", "source_medium[]"),
               ("lecture_context", "lecture{}", "Key is lecture id"),
               ("course_context", "course{}", "Key is course id"),
           ])
@api_moderator_route()
def api_route_get_source_media():
    entries_per_page = api_request_get_query_int("entries_per_page", 20, 10, 50)
    page_num = api_request_get_query_int("page", 0)
    include_sorted = api_request_get_query_boolean("include_sorted", False)
    
    def _trans(session: SessionDb):
        query = SourceMedium.select(is_mod=True, to_load=[
            LOAD_SOURCE_MEDIUM_LECTURE,
            LOAD_LECTURE_COURSE
        ])
        
        if not include_sorted:
            query = query.where(SourceMedium.lecture_id.is_(None))
        
        entries = session.scalars(
            query.offset(page_num * entries_per_page)
            .limit(entries_per_page)
            .order_by(SourceMedium.lecture_id.is_(None).desc(), SourceMedium.update_time.desc())
        ).all()
        count = session.scalar(
            query.with_only_columns(sql.func.count()).select_from(SourceMedium)
        )
        
        course_context = {}
        lecture_context = {}
        
        page = []
        for entry in entries:
            page.append(entry.serialize(
                is_mod=True,
                course_context=course_context,
                lecture_context=lecture_context,
                include_lectures=False,
                include_chapters=False,
                include_media=False,
            ))
        
        session.rollback()
        return {
            "page_count": math.ceil(count / entries_per_page),
            "page": page,
            "lecture_context": lecture_context,
            "course_context": course_context,
        }
    
    return database.execute_read_transaction(_trans)


@api_route("/run_source_file_sorter", "POST",
           response_objects=[
               ("scheduled", "boolean", "If false, another sorter is probably still running"),
           ])
@api_moderator_route(require_csrf_token=True)
def api_route_run_source_file_sorter():
    def trans(session: SessionDb):
        if session.scalar(
                Job.select(is_mod=True, to_load=[])
                   .where(Job.type == "source_file_sorter")
                   .where(Job.status.in_([JobState.READY, JobState.SPAWNING, JobState.RUNNING]))
                   .with_only_columns(sql.func.count())
        ) > 0:
            return False
        session.add(Job(
            type="source_file_sorter",
            input_data={},
            cause_user_id=get_user_id()
        ))
        return True
    
    return {
        "scheduled": database.execute_write_transaction_and_commit(trans)
    }


@api_route("/media_process_templates", "GET", allow_while_readonly=True,
           response_objects=[
               ("templates", "media_process_template[]"),
           ])
@api_moderator_route()
def api_route_get_media_process_templates():
    return {
        "templates": [
            template.serialize(is_mod=True)
            for template in database.query_all_and_expunge(
                MediaProcessTemplate.select(is_mod=True, to_load=[])
            )
        ]
    }


@api_route("/lecture/<int:lecture_id>/media_process_overview", "GET", allow_while_readonly=True,
           response_objects=[
               ("status", "string", "String with some information about the process (e.g. error messages)"),
               ("source_media", "source_medium{}", "Key is id"),
               ("target_media", "target_medium{}", "Key is id. Target media are serialized in extended mode"),
               ("publish_media", "publish_medium{}", "Key is id. Note that these contain their full target media as an "
                                                     "object and not just the id (So there is some duplicate data)"),
           ])
@api_moderator_route()
def api_route_get_lecture_media_process_overview(lecture_id: int):
    lecture = database.query_one_or_none_and_expunge(Lecture.select(is_mod=True, to_load=[
        LOAD_LECTURE_SOURCE_MEDIA,
        LOAD_LECTURE_TARGET_MEDIA,
        LOAD_LECTURE_PUBLISH_MEDIA,
        LOAD_SOURCE_MEDIUM_LECTURE,
        LOAD_PUBLISH_MEDIUM_LECTURE,
        LOAD_PUBLISH_MEDIUM_TARGET_MEDIUM,
        LOAD_TARGET_MEDIUM_LECTURE,
        LOAD_TARGET_MEDIUM_PRODUCER_JOB,
    ]).where(Lecture.id == lecture_id))
    if lecture is None:
        raise ApiClientException(ERROR_UNKNOWN_OBJECT)
    
    source_medium_context = {}
    target_medium_context = {}
    publish_medium_context = {}
    
    for med in chain(lecture.source_media, lecture.target_media, lecture.publish_media):
        med.serialize_to_context(
            is_mod=True,
            extended_target_medium=True,
            source_medium_context=source_medium_context,
            target_medium_context=target_medium_context,
            publish_medium_context=publish_medium_context,
            lecture_context=NoSerializationContext(),
            job_context=NoSerializationContext(),
        )
    return {
        "status": lecture.current_media_process_error,
        "source_media": source_medium_context,
        "target_media": target_medium_context,
        "publish_media": publish_medium_context,
    }


@api_route("/lecture/<int:lecture_id>/run_media_process_scheduler", "POST",
           response_objects=[
               ("scheduled", "boolean", "If false, another scheduler is probably still running"),
           ])
@api_moderator_route(require_csrf_token=True)
def api_route_run_media_process_scheduler(lecture_id: int):
    def trans(session: SessionDb):
        if session.scalar(
            Job.select(is_mod=True, to_load=[])
            .where(Job.type == "media_process_scheduler")
            .where(Job.status.in_([JobState.READY, JobState.SPAWNING, JobState.RUNNING]))
            .where(Job.input_data["lecture_id"].astext.cast(sql.Integer) == lecture_id)
            .with_only_columns(sql.func.count())
        ) > 0:
            return False
        session.add(Job(
            type="media_process_scheduler",
            input_data={
                "lecture_id": lecture_id,
                "manually_triggered": True
            },
            cause_user_id=get_user_id()
        ))
        return True
    
    return {
        "scheduled": database.execute_write_transaction_and_commit(trans)
    }

