from functools import wraps
import traceback
from typing import Callable

from flask import Response, request, Request
from re import Pattern

import api
import app
from api.authentication import is_moderator
from api.miscellaneous import *
from api.version import get_api_path, API_LATEST_VERSION, API_OLDEST_ACTIVE_VERSION
from videoag_common.database.database import TransactionConflictError
from videoag_common.miscellaneous.json import CJsonException

_API_GLOBAL_RATE_LIMITERS = create_configured_host_rate_limiters("global", api.config["API_GLOBAL_RATE_LIMIT"])
_DEFAULT_CACHE_CONTROL_MAX_AGE_SECONDS = api.config["DEFAULT_CACHE_CONTROL_MAX_AGE_SECONDS"]


_SERVER_NAME = api.config["API_SERVER_NAME"]


_LOG_CLIENT_ERRORS: bool = DEBUG_ENABLED and api.config.get("LOG_CLIENT_ERRORS_WHEN_DEBUGGING", True)


def get_client_json(request: Request) -> CJsonObject:
    json = request.get_json(force=False, silent=True, cache=False)
    if json is None:
        raise ApiClientException(ERROR_MALFORMED_JSON)
    return CJsonObject(json)


def api_on_error(error: ApiError):
    error_json = {
        "error_code": error.error_code,
        "message": error.message,
        "server": _SERVER_NAME
    }
    return Response(json.dumps(error_json), error.http_status_code, mimetype="application/json")


def check_client_int(value: int, name: str, min_value: int = MIN_VALUE_UINT32, max_value: int = MAX_VALUE_UINT32):
    if value < min_value:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
            name, f"Value must not be less than {min_value}"))
    if value > max_value:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
            name, f"Value must not be greater than {max_value}"))


def api_request_get_query_string(id: str, max_length: int, pattern: Pattern or None = None, default: str or None = None) -> str or None:
    if id not in request.args:
        return default
    value = request.args[id]
    if len(value) > max_length:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
            f"URL.{id}", f"Must not be longer than {max_length} characters"))
    if pattern is not None and pattern.fullmatch(value) is None:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
            f"URL.{id}", f"Does not match pattern"))
    return value


def api_request_get_query_int(id: str, default: int or None, min_value: int = MIN_VALUE_UINT32, max_value: int = MAX_VALUE_UINT32) -> int or None:
    if id not in request.args:
        return default
    try:
        value = int(request.args[id])
    except ValueError:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
            f"URL.{id}", "Cannot parse integer"))
    check_client_int(value, f"URL.{id}", min_value, max_value)
    return value


def api_request_get_query_boolean(id: str, default: bool or None) -> int or None:
    if id not in request.args:
        return default
    val_str = request.args[id]
    if val_str.lower() == "true":
        return True
    elif val_str.lower() == "false":
        return False
    else:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
            f"URL.{id}", "Cannot parse boolean"
        ))


@api.app.errorhandler(400)
def _handle_bad_request(e=None):
    if _LOG_CLIENT_ERRORS and e is not None:
        print(f"A client exception occurred while handling api request '{request.path}':")
        traceback.print_exception(e)
    return api_on_error(ERROR_BAD_REQUEST("Bad request"))


@api.app.errorhandler(404)
def _handle_not_found(e=None):
    if _LOG_CLIENT_ERRORS and e is not None:
        print(f"A client exception occurred while handling api request '{request.path}':")
        traceback.print_exception(e)
    return api_on_error(ERROR_UNKNOWN_REQUEST_PATH)


@api.app.errorhandler(405)
def _handle_method_not_allowed(e=None):
    if _LOG_CLIENT_ERRORS and e is not None:
        print(f"A client exception occurred while handling api request '{request.path}':")
        traceback.print_exception(e)
    return api_on_error(ERROR_METHOD_NOT_ALLOWED)


@api.app.errorhandler(500)
@api.app.errorhandler(Exception)
def _handle_internal_server_error(e=None):
    if e is not None:
        print(f"An exception occurred while handling api request '{truncate_string(request.path, 200)}':")
        traceback.print_exception(e)
    return api_on_error(ERROR_INTERNAL_SERVER_ERROR)


class ApiResponse:
    
    def __init__(self,
                 data,
                 status: int = HTTP_200_OK,
                 mime_type: str or None = "application/json",
                 headers: dict[str, str] or None = None,
                 default_cache_control_max_age_sec: int or None = _DEFAULT_CACHE_CONTROL_MAX_AGE_SECONDS):
        super().__init__()
        if mime_type == "application/json" and (isinstance(data, dict) or isinstance(data, list)):
            data = json.dumps(data)
        self._default_cache_control_max_age_sec = default_cache_control_max_age_sec
        
        self.response = Response(
            data,
            status=status,
            headers=headers,
            mimetype=mime_type
        )
    
    def build_response(self):
        if "Cache-Control" not in self.response.headers:
            cache_params: list[str] = []
            if (self.response.status_code < 400
                    and request.method == "GET"
                    and "Set-Cookie" not in self.response.headers
                    and not is_moderator()
                    and self._default_cache_control_max_age_sec is not None):
                cache_params.append(f"max-age={self._default_cache_control_max_age_sec}")
                cache_params.append("public")
            else:
                cache_params.append("no-store")
            self.response.headers["Cache-Control"] = ",".join(cache_params)
        
        return self.response


# 2 Values: field id, type
# 3 Values: field id, type, notes
# 4 Values: object variant, field id, type, notes
FieldDescription = tuple[str or None, str, str, str] or tuple[str, str, str] or tuple[str, str]
ObjectDescription = dict[str, list[FieldDescription]] or list[FieldDescription]


class ApiFunction:
    
    def __init__(self,
                 py_func: Callable,
                 allow_while_readonly: bool,
                 allow_while_disabled: bool,
                 url_parameters: list[FieldDescription] or None,
                 request_objects: ObjectDescription or None,
                 description: str,
                 response_description: str,
                 response_object_id: str or None,
                 response_objects: ObjectDescription or None,
                 no_documentation: bool,
                 only_mod: bool,
                 requires_csrf_token: bool,
                 ):
        super().__init__()
        self.py_func = py_func
        self.allow_while_readonly = allow_while_readonly
        self.allow_while_disabled = allow_while_disabled
        self.url_parameters = url_parameters
        self.request_objects = request_objects
        self.description = description
        self.response_description = response_description
        self.response_object_id = response_object_id
        self.response_objects = response_objects
        self.no_documentation = no_documentation
        self.only_mod = only_mod
        self.requires_csrf_token = requires_csrf_token
        
        self.routes: list[ApiRoute] = []


class ApiRoute:
    
    def __init__(self, path: str, full_path: str, method: str, function: ApiFunction):
        super().__init__()
        self.path = path
        self.full_path = full_path
        self.method = method
        self.function = function
        
        function.routes.append(self)
        
        if method not in ["GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", "TRACE", "PATCH"]:
            raise ValueError(f"Unknown http method: {method}")


API_ROUTES_BY_PATH_BY_METHOD: dict[str, dict[str, ApiRoute]] = {}
API_FUNCTIONS: list[ApiFunction] = []
API_NON_STORED_OBJECTS: dict[str, ObjectDescription] = {}
__routes_registered = False


def api_register_non_stored_object(id: str, description: ObjectDescription):
    if id in API_NON_STORED_OBJECTS:
        raise ValueError(f"An object with id '{id}' already exists")
    API_NON_STORED_OBJECTS[id] = description


def __register_routes_to_flask():
    global __routes_registered
    if __routes_registered:
        raise AssertionError()
    __routes_registered = True
    for full_path, route_by_method in API_ROUTES_BY_PATH_BY_METHOD.items():
        assert len(route_by_method) > 0
        if len(route_by_method) == 1:
            route = next(iter(route_by_method.values()))
            app.app.add_url_rule(
                full_path,
                view_func=route.function.py_func,
                methods=[route.method]
            )
            continue
        
        # noinspection PyDefaultArgument
        def _method_handler(__route_by_method=route_by_method, *args, **kwargs):  # Prevent late binding
            return __route_by_method[request.method].function.py_func(*args, **kwargs)
        
        app.app.add_url_rule(
            full_path,
            endpoint=full_path,
            view_func=_method_handler,
            methods=list(route_by_method.keys()),
        )
    
    for func in API_FUNCTIONS:
        if len(func.routes) == 0:
            print(f"WARNING: Api function without any routes: {func.py_func}")


def api_route(path: str, method: str,
              allow_while_readonly: bool = False,
              allow_while_disabled: bool = False,
              min_api_version: int = API_OLDEST_ACTIVE_VERSION,
              max_api_version: int = API_LATEST_VERSION,
              url_parameters: list[FieldDescription] or None = None,
              request_objects: ObjectDescription or None = None,
              description: str = "",
              response_description: str = "",
              response_object_id: str or None = None,
              response_objects: ObjectDescription or None = None,
              no_documentation: bool = False,
              ):
    def decorator(func):
        func = api_function(
            allow_while_readonly=allow_while_readonly,
            allow_while_disabled=allow_while_disabled,
            url_parameters=url_parameters,
            request_objects=request_objects,
            description=description,
            response_description=response_description,
            response_object_id=response_object_id,
            response_objects=response_objects,
            no_documentation=no_documentation
        )(func)
        func = api_add_route(path, method, min_api_version, max_api_version)(func)
        return func
    
    return decorator


def api_add_route(path: str, method: str,
                  min_api_version: int = API_OLDEST_ACTIVE_VERSION,
                  max_api_version: int = API_LATEST_VERSION):
    def decorator(func):
        if not hasattr(func, "api_function"):
            raise Exception("@api_add_route() seems to be applied before @api_function()")
        
        api_function = func.api_function
        
        for version in range(min_api_version, max_api_version + 1):
            full_path = get_api_path(version, path)
            if DEBUG_ENABLED:
                print(f"Registering api route: {full_path}")
            if __routes_registered:
                raise Exception("Api routes have already been registered with flask. Make sure your route is loaded by "
                                "routes/__init__.py before the registration")
            route = ApiRoute(
                path,
                full_path,
                method,
                api_function
            )
            by_method = API_ROUTES_BY_PATH_BY_METHOD.setdefault(full_path, {})
            if method in by_method:
                raise ValueError(f"Duplicate route function for method '{method}' and path '{full_path}'")
            by_method[method] = route
        return func
    
    return decorator


def api_function(track_in_diagnostics: bool = True,
                 allow_while_readonly: bool = False,
                 allow_while_disabled: bool = False,
                 rate_limiters: tuple[HostBasedCounterRateLimiter, ...] = _API_GLOBAL_RATE_LIMITERS,
                 url_parameters: list[FieldDescription] or None = None,
                 request_objects: ObjectDescription or None = None,
                 description: str = "",
                 response_description: str = "",
                 response_object_id: str or None = None,
                 response_objects: ObjectDescription or None = None,
                 no_documentation: bool = False,
                 ):
    """
    TODO doc
    
    :response_object_id: Id of *global* object!
    
    Local response object should have key "" (the empty string)
    """
    def decorator(func):
        if hasattr(func, "api_function"):
            raise Exception("An @api_function() decorator has already been applied. Are you using multiple @api_route()? "
                            "Use @api_add_route(...)@api_add_route(..)@api_function() instead")
        
        call_counter = None
        if track_in_diagnostics:
            func_name: str = func.__name__
            if not func_name.startswith("api_route_"):  # pragma: no cover
                raise RuntimeError("Api route function names must start with 'api_route_' "
                                   "(These names are used in diagnostics data)")
            func_name = func_name[len("api_route_"):]
            if len(func_name) == 0:
                raise RuntimeError("Api route function has no name (just 'api_route_')")  # pragma: no cover
            
            call_counter = DIAGNOSTICS_TRACKER.register_counter(f"route.{func_name}")
        
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                if api.live_config.is_disabled() and not allow_while_disabled:
                    raise ApiClientException(ERROR_SITE_IS_DISABLED)
                if api.live_config.is_readonly() and not allow_while_readonly:
                    raise ApiClientException(ERROR_SITE_IS_READONLY)
                
                if call_counter is not None:
                    call_counter.trigger()
                
                if "X_REAL_IP" in request.headers:
                    ip_string = request.headers["X_REAL_IP"]
                    for rate_limiter in rate_limiters:
                        if not rate_limiter.check_new_request(ip_string):
                            raise ApiClientException(ERROR_RATE_LIMITED)
                
                if DEBUG_ENABLED and "API_ROULETTE_MODE" in api.config:
                    import random
                    from videoag_common.miscellaneous.errors import ALL_ERRORS_RANDOM
                    if random.random() * 100 < int(api.config["API_ROULETTE_MODE"]):
                        raise ApiClientException(random.choice(ALL_ERRORS_RANDOM))
                
                result = func(*args, **kwargs)
                
                if isinstance(result, Response):
                    return result
                elif isinstance(result, ApiResponse):
                    return result.build_response()
                elif result is None:
                    return ApiResponse(None, HTTP_200_OK, None).build_response()
                elif isinstance(result, dict):
                    return ApiResponse(result).build_response()
                elif isinstance(result, tuple) and len(result) == 2:
                    return ApiResponse(result[0], result[1]).build_response()
                else:  # pragma: no cover
                    raise Exception(f"Api route {truncate_string(request.path)} returned result of unknown type: {str(result)}")
            except CJsonException as e:
                if _LOG_CLIENT_ERRORS:
                    print(f"A client exception occurred while handling api request '{request.path}':")
                    traceback.print_exception(e)
                return api_on_error(ERROR_REQUEST_INVALID_PARAMETER(e.path, e.message))
            except ApiClientException as e:
                if _LOG_CLIENT_ERRORS:
                    print(f"A client exception occurred while handling api request '{request.path}':")
                    traceback.print_exception(e)
                return api_on_error(e.error)
            except TransactionConflictError as e:
                print(f"An transaction conflict occurred while handling api request '{truncate_string(request.path, 200)}':")
                traceback.print_exception(e)
                return api_on_error(ERROR_SITE_IS_OVERLOADED)
            except Exception as e:
                print(f"An exception occurred while handling api request '{truncate_string(request.path, 200)}':")
                traceback.print_exception(e)
                return api_on_error(ERROR_INTERNAL_SERVER_ERROR)

        api_function = ApiFunction(
            wrapper,
            allow_while_readonly,
            allow_while_disabled,
            url_parameters,
            request_objects,
            description,
            response_description,
            response_object_id,
            response_objects,
            no_documentation,
            False if not hasattr(func, "api_function_only_mod") else func.api_function_only_mod,
            False if not hasattr(func, "api_function_requires_csrf_token") else func.api_function_requires_csrf_token,
        )
        API_FUNCTIONS.append(api_function)
        
        setattr(wrapper, "api_function", api_function)
        return wrapper
    
    return decorator
