import os
from datetime import date, timedelta

from videoag_common.objects import *
from api.routes import *


_GIT_COMMIT_HASH = os.environ.get("VIDEOAG_API_GIT_COMMIT_HASH", None)
if _GIT_COMMIT_HASH is not None and len(_GIT_COMMIT_HASH) == 0:
    _GIT_COMMIT_HASH = None

_SERVER_NAME = api.config["API_SERVER_NAME"]


@api_route("/status", "GET", allow_while_readonly=True, allow_while_disabled=True,
           response_objects=[
               ("status", "string", "possible values: `available`, `readonly`, `unavailable`"),
               ("is_debug", "boolean"),
               ("git_commit_hash", "?string"),
               ("server_name", "string"),
               ("announcements", "announcement[]")
           ])
def api_route_status():
    is_mod = is_moderator()
    status = "available"
    announcements = []
    announcements.extend(api.live_config.get_static_announcements())

    if api.live_config.is_disabled():
        status = "unavailable"
    else:
        if api.live_config.is_readonly():
            status = "readonly"
        # noinspection PyBroadException
        query = Announcement.select(is_mod, to_load=[])
        db_rows = []
        try:
            db_rows = database.query_all_and_expunge(query)
        except Exception:
            # We do NOT return status 'unavailable' because the status is the EXPECTED status. If there is an unexpected
            # db failure, we want errors to show up and not just 'Site not available'
            pass
        announcements.extend(map(
            lambda a: a.serialize(is_mod=is_mod),
            db_rows
        ))
    response = {
        "status": status,
        "is_debug": DEBUG_ENABLED,
        "server_name": _SERVER_NAME,
        "announcements": announcements
    }
    if _GIT_COMMIT_HASH is not None:
        response["git_commit_hash"] = _GIT_COMMIT_HASH
    return response


@api_route("/is_running", "GET", allow_while_readonly=True, allow_while_disabled=True,
           response_description="""
Returns `200 OK` if the site is ready to answer requests, otherwise `503 SERVICE UNAVAILABLE`. The status may still be
`unavailable` if this return `200 OK`!
""")
def api_route_is_running():
    return {}, HTTP_200_OK


def _db_execute_get_homepage(session: SessionDb, is_mod: bool):
    upcoming_start: date = date.today()
    upcoming_end: date = upcoming_start + timedelta(days=7)

    upcoming_lectures = session.scalars(
        Lecture.select(
            is_mod,
            [
                LOAD_LECTURE_COURSE,
                LOAD_LECTURE_CHAPTERS,
                LOAD_LECTURE_PUBLISH_MEDIA,
                LOAD_PUBLISH_MEDIUM_TARGET_MEDIUM
            ],
            visibility_check=True,
        )
        .where(Lecture.time >= upcoming_start, Lecture.time <= upcoming_end)
        .where(~Lecture.no_recording)
        .order_by(Lecture.time.asc(), Lecture.id.asc())
        .limit(30)
    ).all()
    latest_lectures = session.scalars(
        Lecture.select(
            is_mod,
            [
                LOAD_LECTURE_COURSE,
                LOAD_LECTURE_CHAPTERS,
                LOAD_LECTURE_PUBLISH_MEDIA,
                LOAD_PUBLISH_MEDIUM_TARGET_MEDIUM
            ],
            visibility_check=True,
        )
        .where(Lecture.publish_time.is_not(None))
        .order_by(Lecture.publish_time.desc())
        .limit(6)
    ).all()
    featured = session.scalars(
        Featured.select(is_mod, [
            LOAD_FEATURED_COURSE_OR_LECTURE,
            LOAD_LECTURE_COURSE,
            LOAD_LECTURE_CHAPTERS,
            LOAD_LECTURE_PUBLISH_MEDIA,
            LOAD_PUBLISH_MEDIUM_TARGET_MEDIUM
        ])
                .order_by(Featured.display_priority.asc())
    ).all()
    session.expunge_all()
    session.rollback()

    course_context = {}

    return {
        "featured": list(map(lambda f: f.serialize(
            is_mod=is_mod,
            include_lectures=False,
            include_chapters=True,
            include_media=True,
            course_context=course_context
        ), featured)),
        "upcoming_lectures": list(map(lambda l: l.serialize(
            is_mod=is_mod,
            include_lectures=False,
            include_chapters=True,
            include_media=True,
            course_context=course_context
        ), upcoming_lectures)),
        "latest_lectures": list(map(lambda l: l.serialize(
            is_mod=is_mod,
            include_lectures=False,
            include_chapters=True,
            include_media=True,
            course_context=course_context
        ), latest_lectures)),
        "course_context": course_context
    }


@api_route("/homepage", "GET", allow_while_readonly=True,
           response_objects=[
               ("featured", "featured[]"),
               ("upcoming_lectures", "lecture[]", "Soonest is first"),
               ("latest_lectures", "lecture[]", "Latest is first"),
               ("course_context", "course{}", "Key is course id. Contains all courses referenced in the lectures")
           ])
def api_route_homepage():
    is_mod: bool = is_moderator()

    return database.execute_read_transaction(
        _db_execute_get_homepage,
        is_mod
    )
