import flask

from videoag_common.objects import *

from api.routes import *
from api.settings import Settings, SettingsSection, SettingsBooleanEntry, SettingsSubSectionEntry


_SETTINGS = Settings([
    SettingsSection("notifications", [
        SettingsBooleanEntry("notifications_enabled", "enable_mail_notifications"),
        SettingsSubSectionEntry("notifications_responsible_courses", [
            SettingsBooleanEntry("notifications_responsible_courses_new_media_source", "notify_new_video"),
            SettingsBooleanEntry("notifications_responsible_courses_chapter_submitted", "notify_chapter_suggested"),
            SettingsBooleanEntry("notifications_responsible_courses_changes", "notify_edit"),
        ])
    ])
])


api_register_non_stored_object("settings", [
    ("sections", "settings_section[]")
])

api_register_non_stored_object("settings_section", [
    ("name", "string", "Lang Key: `setting.section.{name}`"),
    ("entries", "settings_entry[]")
])

api_register_non_stored_object("settings_entry", [
    (None, "id", "string", "Lang Key: `setting.entry.{id}.name` and `setting.entry.{id}.description` (description may not exist)"),
    (None, "type", "string", "Type of this entry. For possible values see this table"),
    ("subsection", "entries", "settings_entry[]", "Contain no subsections"),
    ("boolean", "value", "boolean", "")
])


@api_route("/users", "GET", allow_while_readonly=True,
           response_description="Returns a list of all users (moderators)",
           response_objects=[
               ("users", "user[]")
           ])
@api_moderator_route()
def api_route_users():
    users = database.query_all_and_expunge(User.basic_select())
    return {
        "users": list(map(lambda u: u.serialize(is_mod=True), users))
    }


@api_route("/user/me/settings", "GET", allow_while_readonly=True,
           response_object_id="settings")
@api_moderator_route()
def api_route_get_user_me_settings():
    user = database.query_one_or_none_and_expunge(User.basic_select().where(User.id == get_user_id()))
    if user is None:
        raise RuntimeError("No db user for moderator")  # pragma: no cover
    return _SETTINGS.to_json(user)


@api_route("/user/me/settings", "PATCH",
           request_objects=[
               ("updates", "{}", "Key is id of setting_entry from `GET /user/me/settings` to change (All types except "
                                 "`subsection`). Value is new value of corresponding type")
           ],
           response_description="Updates the specified settings or returns an [api_error](#api_error) if the update failed.")
@api_moderator_route(require_csrf_token=True)
def api_route_patch_user_me_settings():
    updates_json = get_client_json(flask.request).get_object("updates")
    
    def _trans(session: SessionDb):
        user = session.scalar(User.basic_select().where(User.id == get_user_id()))
        if user is None:
            raise RuntimeError("No db user for moderator")  # pragma: no cover
        for entry_id in updates_json.keys():
            entry = _SETTINGS.get_value_entry(entry_id)
            if entry is None:
                raise ApiClientException(ERROR_BAD_REQUEST(f"Unknown user settings id {truncate_string(entry_id)}"))
            entry.set_value(user, entry.json_value_to_db(updates_json.get(entry_id)))
        session.commit()
    
    database.execute_write_transaction(_trans)
    return {}
