from abc import ABC, ABCMeta, abstractmethod
from typing import Generic, TypeVar

from api.miscellaneous import *


_T = TypeVar("_T")


class SettingsEntry(Generic[_T], ABC):
    __metaclass__ = ABCMeta
    
    def __init__(self, id: str, type_id: str):
        super().__init__()
        self.id = id
        self._type_id = type_id
    
    def to_json(self, db_object: _T) -> dict:
        return {
            "id": self.id,
            "type": self._type_id
        }


class SettingsValueEntry(Generic[_T], SettingsEntry[_T], ABC):
    __metaclass__ = ABCMeta
    
    def __init__(self, id: str, type_id: str, object_attr: str):
        super().__init__(id, type_id)
        self.object_attr = object_attr
    
    def to_json(self, db_object: _T) -> dict:
        return super().to_json(db_object) | {
            "value": self._db_value_to_json(getattr(db_object, self.object_attr))
        }
    
    def set_value(self, db_object: _T, value):
        setattr(db_object, self.object_attr, value)
    
    @abstractmethod
    def _db_value_to_json(self, db_value):
        pass  # pragma: no cover
    
    @abstractmethod
    def json_value_to_db(self, json_value: CJsonValue):
        pass  # pragma: no cover


class SettingsBooleanEntry(Generic[_T], SettingsValueEntry[_T]):
    
    def __init__(self, id: str, db_colum: str):
        super().__init__(id, "boolean", db_colum)
    
    def _db_value_to_json(self, db_value):
        return bool(db_value)
    
    def json_value_to_db(self, json_value: CJsonValue):
        return json_value.as_bool()


class SettingsSubSectionEntry(Generic[_T], SettingsEntry[_T]):
    
    def __init__(self, id: str, entries: list[SettingsValueEntry]):
        super().__init__(id, "subsection")
        self.entries = entries
    
    def to_json(self, db_object: _T) -> dict:
        return super().to_json(db_object) | {
            "entries": list(map(lambda e: e.to_json(db_object), self.entries))
        }


class SettingsSection:
    
    def __init__(self, name_key: str, entries: list[SettingsEntry]):
        super().__init__()
        self._name_key = name_key
        self.entries = entries
    
    def to_json(self, db_object: _T):
        return {
            "name": self._name_key,
            "entries": list(map(lambda e: e.to_json(db_object), self.entries))
        }


class Settings:
    
    def __init__(self, sections: list[SettingsSection]):
        super().__init__()
        self._sections = sections
        self._value_entries_by_id: dict[str, SettingsValueEntry] = {}
        for section in sections:
            self._collect_entries(section.entries)
    
    def _collect_entries(self, entries: list[SettingsEntry]):
        for entry in entries:
            if isinstance(entry, SettingsSubSectionEntry):
                self._collect_entries(entry.entries)
                continue
            assert isinstance(entry, SettingsValueEntry)
            if entry.id in self._value_entries_by_id:
                raise RuntimeError(f"Duplicate settings id {entry.id}")  # pragma: no cover
            self._value_entries_by_id[entry.id] = entry
    
    def get_value_entry(self, id: str) -> SettingsValueEntry or None:
        return self._value_entries_by_id.get(id, None)
    
    def to_json(self, db_object: _T):
        return {
            "sections": list(map(lambda s: s.to_json(db_object), self._sections))
        }
