import re

from videoag_common.objects import API_CLASSES_BY_ID, ApiObjectClass
from videoag_common.api_object import ApiConfigField

# One needs to be careful with lists. As the generated json might be used for a diff, lists should always be sorted

_TYPE_PATTERN = re.compile("[a-zA-Z0-9_]+")

_STANDARD_TYPES = ["string", "int", "boolean", "datetime", "json", "enum"]


_PATTERN_MARKDOWN_CODE = re.compile("`(.*?)`")

# TODO nullability


def _markdown_to_html(value: str) -> str:
    res = value
    res = _PATTERN_MARKDOWN_CODE.sub("<code>\\1</code>", res)
    # TODO link, more
    return res


def _check_type(specification: "ApiSpecification", local_objects: dict[str, "ObjectSpecification"], type: str):
    if type.startswith("NOCHECK:"):
        return
    if type == "{}":
        return
    if type.endswith("[]"):
        type = type[:len(type) - 2]
    elif type.endswith("{}"):
        type = type[:len(type) - 2]
    
    if not _TYPE_PATTERN.fullmatch(type):
        raise ValueError(f"Type contains invalid characters: '{type}'")
    
    if type in local_objects:
        return
    
    if type in specification.global_objects:
        return
    
    if type in _STANDARD_TYPES:
        return
    
    raise ValueError(f"Unknown type: '{type}'")


class FieldSpecification:
    
    def __init__(self,
                 id: str,
                 type: str,
                 check_type: bool,
                 object_variant: str or None,
                 notes: str,
                 only_mod: bool,
                 optional: bool,
                 config_directly_modifiable: bool
                 ):
        super().__init__()
        self.id = id
        self.type = type
        self.check_type = check_type
        self.object_variant = object_variant
        self.notes = notes
        self.only_mod = only_mod
        self.optional = optional
        self.config_directly_modifiable = config_directly_modifiable
    
    def check_valid(self, specification: "ApiSpecification", local_objects: dict[str, "ObjectSpecification"]):
        if self.check_type:
            _check_type(specification, local_objects, self.type)
        pass
    
    def to_json(self):
        return {
            "id": self.id,
            "type": self.type,
            "object_variant": self.object_variant,
            "notes": self.notes,
            "only_mod": self.only_mod,
            "optional": self.optional,
            "config_directly_modifiable": self.config_directly_modifiable
        }
    
    def generate_markdown_row(self, variant_row_span: int or None) -> str:
        type = self.type
        if self.only_mod:
            type = "!" + type
        if self.optional:
            type = "?" + type
        if self.config_directly_modifiable:
            type = "+" + type
        
        res = "    <tr>\n"
        if variant_row_span is not None:
            if self.object_variant is None:
                own_variant = "any"
            elif self.object_variant.startswith("NOCODE:"):
                own_variant = self.object_variant[len("NOCODE:"):]
            else:
                own_variant = f"`{self.object_variant}`"
            res += f"        <td rowspan=\"{variant_row_span}\">{_markdown_to_html(own_variant)}</td>\n"
        
        res += f"        <td>{_markdown_to_html(self.id)}</td>\n"
        res += f"        <td>{_markdown_to_html(type)}</td>\n"   # TODO link
        res += f"        <td>{_markdown_to_html(self.notes)}</td>\n"
        
        res += "    </tr>\n"
        return res


class ObjectSpecification:
    
    def __init__(self,
                 id: str or None,
                 fields_by_variant_by_id: dict[str or None, dict[str, FieldSpecification]],
                 ):
        super().__init__()
        self.id = id
        self.fields_by_variant_by_id = fields_by_variant_by_id
    
    def check_valid(self, specification: "ApiSpecification", local_objects: dict[str, "ObjectSpecification"]):
        for variant, fields_by_id in self.fields_by_variant_by_id.items():
            for field in fields_by_id.values():
                try:
                    field.check_valid(specification, local_objects)
                    if variant is not None and field.id in self.fields_by_variant_by_id[None]:
                        raise ValueError("Field has same id as non-variant field")
                except Exception as e:
                    raise Exception(f"In field '{field.id}' for variant '{variant}'") from e
    
    def to_json(self):
        return {
            "id": self.id,
            "fields": {
                "" if variant is None else variant: {
                    id: field.to_json()
                    for id, field in fields.items()
                }
                for variant, fields in self.fields_by_variant_by_id.items()
            }
        }
    
    def generate_markdown(self, field_id_title: str = "Field") -> str:
        res = "<table>\n"
        has_variants = len(self.fields_by_variant_by_id) > 1
        
        res += "<thead>\n"
        if has_variants:
            res += "    <th>Variant</th>\n"
        res += f"    <th>{field_id_title}</th>\n"
        res += f"    <th>Type</th>\n"
        res += f"    <th>Notes</th>\n"
        res += "</thead>\n"
        
        res += "<tbody>\n"
        for variant in [None] + list(sorted(filter(lambda v: v is not None, self.fields_by_variant_by_id.keys()))):
            fields_by_id = self.fields_by_variant_by_id[variant]
            row_span = None
            if len(self.fields_by_variant_by_id) > 1:
                row_span = len(fields_by_id)
            for field_id in sorted(fields_by_id.keys()):
                res += fields_by_id[field_id].generate_markdown_row(row_span)
                row_span = None
            
            if variant is not None and len(fields_by_id) == 0:
                assert isinstance(variant, str)
                if variant.startswith("NOCODE:"):
                    variant_str = variant[len("NOCODE:"):]
                else:
                    variant_str = f"`{variant}`"
                res += f"    <tr><td>{_markdown_to_html(variant_str)}</td></tr>\n"
        res += "</tbody>\n"
        
        res += "</table>\n\n"
        return res


_ROUTE_PARAMETER_PATTERN = re.compile("<(.+?:)?(.+?)>")


class RouteSpecification:
    
    def __init__(self,
                 method: str,
                 path: str):
        super().__init__()
        self.method = method
        self.path = path
        
        # We ignore the parameter type
        self.parameter_ids = map(lambda match: match.group(2), _ROUTE_PARAMETER_PATTERN.finditer(path))
        self.display_path = _ROUTE_PARAMETER_PATTERN.sub("{\\2}", path)
        self.identifier = self.path + ":" + self.method
    
    def to_json(self):
        return {
            "method": self.method,
            "path": self.path,
            "display_path": self.display_path
        }


class FunctionSpecification:
    
    def __init__(self,
                 routes: list[RouteSpecification],
                 allow_while_readonly: bool,
                 allow_while_disabled: bool,
                 url_parameters: ObjectSpecification or None,
                 request_objects: dict[str, ObjectSpecification],
                 description: str,
                 response_description: str,
                 response_object_id: str,
                 response_objects: dict[str, ObjectSpecification],
                 only_mod: bool,
                 requires_csrf_token: bool,
                 ):
        super().__init__()
        self.routes = sorted(routes, key=lambda route: route.identifier)
        self.allow_while_readonly = allow_while_readonly
        self.allow_while_disabled = allow_while_disabled
        self.url_parameters = url_parameters
        self.request_objects = request_objects
        self.description = description
        self.response_description = response_description
        self.response_object_id = response_object_id
        self.response_objects = response_objects
        self.only_mod = only_mod
        self.requires_csrf_token = requires_csrf_token
        
        self.identifier = ";".join(map(lambda route: route.identifier, self.routes))
    
    def check_valid(self, specification: "ApiSpecification"):
        
        if len(self.request_objects) > 0 and "" not in self.request_objects:
            raise ValueError("Missing root request object (with id \"\" (empty string))")
        
        if len(self.response_objects) > 0 and "" not in self.response_objects:
            raise ValueError("Missing root response object (with id \"\" (empty string))")
        
        if self.response_object_id == "" and "" not in self.response_objects:
            raise ValueError("Missing root response object (with id \"\" (empty string))")
        
        for obj in self.request_objects.values():
            try:
                obj.check_valid(specification, self.request_objects)
            except Exception as e:
                raise Exception(f"In request object '{obj.id}'") from e
        
        for obj in self.response_objects.values():
            try:
                obj.check_valid(specification, self.response_objects)
            except Exception as e:
                raise Exception(f"In response object '{obj.id}'") from e
        
        if self.response_object_id is not None and self.response_object_id != "":
            try:
                _check_type(specification, self.response_objects, self.response_object_id)
            except Exception as e:
                raise Exception(f"For response type id") from e
    
    def to_json(self):
        return {
            "routes": [route.to_json() for route in self.routes],
            "allow_while_readonly": self.allow_while_readonly,
            "allow_while_disabled": self.allow_while_disabled,
            "url_parameters": None if self.url_parameters is None else self.url_parameters.to_json(),
            "request_objects": {
                id: obj.to_json()
                for id, obj in self.request_objects.items()
            },
            "description": self.description,
            "response_description": self.response_description,
            "response_object_id": self.response_object_id,
            "response_objects": {
                id: obj.to_json()
                for id, obj in self.response_objects.items()
            },
            "only_mod": self.only_mod,
            "requires_csrf_token": self.requires_csrf_token,
        }
    
    def generate_markdown(self) -> str:
        res = ""
        
        res += f"#### `{self.routes[0].method} {self.routes[0].display_path}`\n"
        for route in self.routes[1:]:
            res += f"`{route.method} {route.display_path}`\n"
        res += "\n"
        
        if self.allow_while_readonly and self.allow_while_disabled:
            res += "This may be used while the site is [read-only or disabled](#site-status).\n\n"
        elif self.allow_while_readonly:
            res += "This may be used while the site is [read-only](#site-status).\n\n"
        
        if self.only_mod and self.requires_csrf_token:
            res += "This may only be used by a moderator and requires the CSRF Token.\n\n"
        elif self.only_mod:
            res += "This may only be used by a moderator.\n\n"
        
        if self.url_parameters is not None:
            res += self.url_parameters.generate_markdown(field_id_title="URL Parameter")
            res += "\n"
        
        if len(self.request_objects) > 0:
            for object_id in [""] + list(sorted(filter(lambda v: v != "", self.request_objects.keys()))):
                if object_id != "":
                    res += f"{object_id}:\n"
                res += self.request_objects[object_id].generate_markdown()
                res += "\n"
        
        if len(self.description) > 0:
            res += self.description
            res += "\n\n"
        
        res += "###### Response:\n\n"
        
        if self.response_object_id is not None and self.response_object_id != "":
            res += f"Returns {self.response_object_id}\n\n"  # TODO link
        
        if len(self.response_objects) > 0:
            for object_id in [""] + list(sorted(filter(lambda v: v != "", self.response_objects.keys()))):
                if object_id != "":
                    res += f"{object_id}:\n"
                res += self.response_objects[object_id].generate_markdown()
                res += "\n"
        
        if len(self.response_description) > 0:
            res += self.response_description
            res += "\n\n"
        
        res += "\n---\n"
        
        return res


class ApiSpecification:
    
    def __init__(self,
                 functions: list[FunctionSpecification],
                 global_objects: dict[str, ObjectSpecification],
                 config_used_types: list[str],
                 config_object_ids: list[str],
                 ):
        super().__init__()
        self.functions = functions
        self.global_objects = global_objects
        self.config_used_types = list(sorted(config_used_types))
        self.config_object_ids = list(sorted(config_object_ids))
    
    def check_valid(self):
        for function in self.functions:
            try:
                function.check_valid(self)
            except Exception as e:
                raise Exception(f"In function {function.identifier}") from e
        for obj in self.global_objects.values():
            try:
                obj.check_valid(self, {})
            except Exception as e:
                raise Exception(f"In object {obj.id}") from e
        for type in self.config_used_types:
            try:
                _check_type(self, {}, type)
            except Exception as e:
                raise Exception(f"For type used by a config field: {type}") from e
    
    def to_json(self):
        return {
            "functions": [func.to_json() for func in self.functions],
            "global_objects": {
                key: spec.to_json()
                for key, spec in self.global_objects.items()
            },
            "config_used_types": self.config_used_types,
            "config_object_ids": self.config_object_ids
        }

    def generate_markdown(self, template: str) -> str:
        res = ""
        for function in self.functions:
            res += function.generate_markdown()
        template = template.replace("{{ROUTES_SPECIFICATION}}", res)
        
        res = ""
        for object_id in sorted(self.global_objects.keys()):
            res += f"#### {object_id}\n\n"
            res += self.global_objects[object_id].generate_markdown()
        template = template.replace("{{OBJECTS_SPECIFICATION}}", res)
        
        res = ""
        for type in self.config_object_ids:
            res += f"* {type}\n"
        template = template.replace("{{CONFIG_OBJECT_LIST}}", res)
        
        res = ""
        for type in self.config_used_types:
            res += f"* {type}\n"
        template = template.replace("{{CONFIG_FIELD_USED_TYPE_LIST}}", res)
        
        return template


def _extract_object_specification(id: str or None, fields: list[tuple[str, str, str] or tuple[str, str]]) -> ObjectSpecification:
    field_specs_by_variant_by_id: dict[str or None, dict[str, FieldSpecification]] = {}
    
    for field in fields:
        field_variant = None
        note = ""
        
        if len(field) == 4:
            field_variant, field_id, type, note = field
        elif len(field) == 3:
            field_id, type, note = field
        elif len(field) == 2:
            field_id, type = field
        else:
            raise ValueError("Invalid amount of values in tuple")
        
        only_mod = False
        optional = False
        check_type = True
        
        for _ in range(2):  # Allow any order of modifiers
            if type.startswith("!"):
                if only_mod:
                    raise ValueError(f"Duplicate modifier '!' for field '{field_id}' in object '{id}'")
                only_mod = True
                type = type[1:]
            if type.startswith("?"):
                if optional:
                    raise ValueError(f"Duplicate modifier '?' for field '{field_id}' in object '{id}'")
                optional = True
                type = type[1:]
        
        if type.startswith("NOCHECK:"):
            check_type = False
            type = type[len("NOCHECK:"):]
        
        field_spec = FieldSpecification(
            field_id,
            type,
            check_type,
            field_variant,
            note,
            only_mod=only_mod,
            optional=optional,
            config_directly_modifiable=False
        )
        sub_fields = field_specs_by_variant_by_id.setdefault(field_variant, {})
        if field_id in sub_fields:
            raise ValueError(f"Duplicate field with id '{field_id}' for variant '{field_variant}'")
        sub_fields[field_id] = field_spec
    
    return ObjectSpecification(
        id,
        field_specs_by_variant_by_id
    )


def _extract_api_object(clazz: ApiObjectClass) -> ObjectSpecification or None:
    if not clazz.enable_data:
        return None

    field_specs_by_variant_by_id: dict[str or None, dict[str, FieldSpecification]] = {
        None: {}
    }
    
    if clazz.get_variants() is not None:
        for variant in clazz.get_variants():
            field_specs_by_variant_by_id[variant] = {}
    
    for field in clazz.get_data_fields():
        field_specs_by_variant_by_id[field.variant_id][field.data_id] = FieldSpecification(
            field.data_id,
            field.data_get_type_id(),
            True,
            field.variant_id,
            field.data_notes,
            field.data_only_mod,
            field.data_is_optional(),
            isinstance(field, ApiConfigField) and field.include_in_config and field.config_directly_modifiable
        )
    
    return ObjectSpecification(
        clazz.id,
        field_specs_by_variant_by_id
    )


def _extract_function_objects(objects):
    if objects is None:
        return {}
    
    object_specs: dict[str, ObjectSpecification] = {}
    if isinstance(objects, list):
        object_specs[""] = _extract_object_specification("", objects)
    else:
        assert isinstance(objects, dict)
        for id, fields in objects.items():
            if id in object_specs:
                raise ValueError(f"Duplicate object with id '{id}'")
            object_specs[id] = _extract_object_specification(id, fields)
    return object_specs


def _extract_function(function) -> FunctionSpecification:
    from api.routes.route import ApiFunction
    assert isinstance(function, ApiFunction)
    
    url_parameters: ObjectSpecification or None = None
    if function.url_parameters is not None and len(function.url_parameters) > 0:
        url_parameters = _extract_object_specification(None, function.url_parameters)
    
    request_objects: dict[str, ObjectSpecification] = _extract_function_objects(function.request_objects)
    response_objects: dict[str, ObjectSpecification] = _extract_function_objects(function.response_objects)
    
    response_object_id = function.response_object_id
    if response_object_id is None and "" in response_objects:
        response_object_id = ""
    
    return FunctionSpecification(
        routes=list(map(lambda route: RouteSpecification(route.method, route.path), function.routes)),
        allow_while_readonly=function.allow_while_readonly,
        allow_while_disabled=function.allow_while_disabled,
        url_parameters=url_parameters,
        request_objects=request_objects,
        description=function.description,
        response_description=function.response_description,
        response_object_id=response_object_id,
        response_objects=response_objects,
        only_mod=function.only_mod,
        requires_csrf_token=function.requires_csrf_token
    )


def extract() -> ApiSpecification:
    from api.routes.route import API_FUNCTIONS, API_NON_STORED_OBJECTS
    
    function_specs: list[FunctionSpecification] = []
    for function in API_FUNCTIONS:
        if len(function.routes) == 0 or function.no_documentation:
            continue
        try:
            function_specs.append(_extract_function(function))
        except Exception as e:
            raise Exception(f"In function for path {function.routes[0].path}") from e
    
    global_objects = {}
    for clazz in API_CLASSES_BY_ID.values():
        try:
            object_spec = _extract_api_object(clazz)
        except Exception as e:
            raise Exception(f"In api object {clazz.id}") from e
        
        if object_spec is not None:
            if object_spec.id in global_objects:
                raise ValueError(f"Duplicate global object with id: {object_spec.id}")
            global_objects[object_spec.id] = object_spec
    
    config_used_types: set[str] = set()
    for clazz in API_CLASSES_BY_ID.values():
        if not clazz.enable_config:
            continue
        for field in clazz.get_config_fields():
            config_used_types.add(field.config_get_type_id())
    
    config_object_ids: list[str] = []
    for clazz in API_CLASSES_BY_ID.values():
        if not clazz.enable_config:
            continue
        config_object_ids.append(clazz.id)
    
    for id, desc in API_NON_STORED_OBJECTS.items():
        try:
            object_spec = _extract_object_specification(id, desc)
        except Exception as e:
            raise Exception(f"In non-stored object {id}") from e
        
        if object_spec.id in global_objects:
            raise ValueError(f"Duplicate global object with id: {object_spec.id}")
        global_objects[object_spec.id] = object_spec
    
    spec = ApiSpecification(
        function_specs,
        global_objects,
        list(config_used_types),
        config_object_ids
    )
    spec.check_valid()
    return spec
