from api.database import *
from api.miscellaneous import *
from api.version import API_LATEST_VERSION, get_api_path
from videoag_common.test.database_test import DatabaseTest


class ApiTest(DatabaseTest):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._client = api.app.test_client()
        self._last_request: tuple[str, dict, dict] or None = None
    
    def __call__(self, result=...):
        res = super().__call__(result)
        if not res.wasSuccessful() and self._last_request is not None:
            def pretty_json_or_string(val):
                if val is None:
                    return val
                try:
                    if isinstance(val, dict):
                        json_val = val
                    else:
                        json_val = json.loads(val)
                    return json.dumps(json_val, indent=2)
                except ValueError:
                    return val
            print(f"Last request:{self._last_request[0]}\n"
                  f"{pretty_json_or_string(self._last_request[1])}\n"
                  f"Response:\n"
                  f"{pretty_json_or_string(self._last_request[2])}")
        return res
    
    def moderator_login(self) -> str:
        response = self._client.post(path=f"/api/v{API_LATEST_VERSION}/authentication/fsmpi", json={
            "username": "videoag",
            "password": "videoag"
        })
        for cookie in response.headers.get_all("Set-Cookie"):
            if not cookie.startswith("csrf_token="):
                continue
            return cookie[len("csrf_token="):cookie.index(";")]
        raise RuntimeError("Got no csrf token on login")
    
    def moderator_logout(self):
        self._client.post(path=f"/api/v{API_LATEST_VERSION}/authentication/moderator_logout")
    
    def do_json_request(self,
                        method: str,
                        path: str,
                        request: object or None = None,
                        expected_response: object or None = None,
                        ignore_array_order_keys: list[str] or None = None,
                        ignore_value_keys: list[str] or None = None,
                        assert_empty_response: bool = False,
                        use_moderator_login: bool = False,
                        use_csrf_token: bool = True,
                        expected_response_code: int or None = HTTP_200_OK,
                        api_version: int = API_LATEST_VERSION) -> tuple[int, dict]:
        if not path.startswith("/"):
            raise ValueError("Path must start with /")
        
        headers: list[tuple[str, str]] = []
        
        if use_moderator_login:
            csrf_token = self.moderator_login()
            if use_csrf_token:
                headers.append(("X-Csrf-Token", csrf_token))
        else:
            self.moderator_logout()
        
        if request is None:
            response = self._client.open(
                method=method,
                path=get_api_path(api_version, path),
                headers=headers
            )
        else:
            response = self._client.open(
                method=method,
                path=get_api_path(api_version, path),
                mimetype="application/json",
                data=json.dumps(request),
                headers=headers
            )
        
        data = None
        if response.mimetype == "application/json":
            try:
                data = json.loads(response.get_data(as_text=True))
            except Exception as e:
                if expected_response_code is not None:
                    self.assertEqual(response.status_code, expected_response_code)
                raise AssertionError("Malformed json") from e
        
        self._last_request = (path, request, data)
        
        if expected_response_code is not None:
            if response.status_code != expected_response_code:
                raise AssertionError(f"Expected status {expected_response_code} but got {response.status_code}. "
                                     f"JSON Data: {data}")
        
        if expected_response is not None:
            if assert_empty_response:
                raise ValueError("Cannot assert empty response while expecting a response")
            if data is None:
                raise AssertionError("Expected response but got none (Content-Type is not 'application/json')")
            self.assertJsonEqual(expected_response, data, ignore_array_order_keys, ignore_value_keys)
        elif assert_empty_response and data is not None and len(data) > 0:
            raise AssertionError("Expected no response but got one: " + data)
        
        return response.status_code, data
